#!/usr/bin/python
# Copyright (c) 2021-2022 Hewlett Packard Enterprise, Inc. All rights reserved.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: ilo_redfish_info
short_description: Gathers server information through iLO using Redfish APIs
version_added: 4.2.0
description:
  - Builds Redfish URIs locally and sends them to iLO to get information back.
  - For use with HPE iLO operations that require Redfish OEM extensions.
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module
  - community.general._redfish
options:
  category:
    required: true
    description:
      - List of categories to execute on iLO.
    type: list
    elements: str
  command:
    required: true
    description:
      - List of commands to execute on iLO.
    type: list
    elements: str
  baseuri:
    required: true
    description:
      - Base URI of iLO.
    type: str
  username:
    description:
      - Username for authenticating to iLO.
    type: str
  password:
    description:
      - Password for authenticating to iLO.
    type: str
  auth_token:
    description:
      - Security token for authenticating to iLO.
    type: str
  timeout:
    description:
      - Timeout in seconds for HTTP requests to iLO.
    default: 10
    type: int
  validate_certs:
    version_added: 10.6.0
  ca_path:
    version_added: 10.6.0
  ciphers:
    version_added: 10.6.0
author:
  - "Bhavya B (@bhavya06)"
"""

EXAMPLES = r"""
- name: Get iLO Sessions
  community.general.ilo_redfish_info:
    category: Sessions
    command: GetiLOSessions
    baseuri: "{{ baseuri }}"
    username: "{{ username }}"
    password: "{{ password }}"
  register: result_sessions
"""

RETURN = r"""
ilo_redfish_info:
  description: Returns iLO sessions.
  type: dict
  contains:
    GetiLOSessions:
      description: Returns the iLO session msg and whether the function executed successfully.
      type: dict
      contains:
        ret:
          description: Check variable to see if the information was successfully retrieved.
          type: bool
        msg:
          description: Information of all active iLO sessions.
          type: list
          elements: dict
          contains:
            Description:
              description: Provides a description of the resource.
              type: str
            Id:
              description: The sessionId.
              type: str
            Name:
              description: The name of the resource.
              type: str
            UserName:
              description: Name to use to log in to the management processor.
              type: str
  returned: always
"""

CATEGORY_COMMANDS_ALL = {"Sessions": ["GetiLOSessions"]}

CATEGORY_COMMANDS_DEFAULT = {"Sessions": "GetiLOSessions"}

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ilo_redfish_utils import iLORedfishUtils
from ansible_collections.community.general.plugins.module_utils._redfish_utils import REDFISH_COMMON_ARGUMENT_SPEC


def main():
    result = {}
    category_list = []
    argument_spec = dict(
        category=dict(required=True, type="list", elements="str"),
        command=dict(required=True, type="list", elements="str"),
        baseuri=dict(required=True),
        username=dict(),
        password=dict(no_log=True),
        auth_token=dict(no_log=True),
        timeout=dict(type="int", default=10),
    )
    argument_spec.update(REDFISH_COMMON_ARGUMENT_SPEC)
    module = AnsibleModule(
        argument_spec,
        required_together=[
            ("username", "password"),
        ],
        required_one_of=[
            ("username", "auth_token"),
        ],
        mutually_exclusive=[
            ("username", "auth_token"),
        ],
        supports_check_mode=True,
    )

    creds = {"user": module.params["username"], "pswd": module.params["password"], "token": module.params["auth_token"]}

    timeout = module.params["timeout"]

    root_uri = f"https://{module.params['baseuri']}"
    rf_utils = iLORedfishUtils(creds, root_uri, timeout, module)

    # Build Category list
    if "all" in module.params["category"]:
        for entry in CATEGORY_COMMANDS_ALL:
            category_list.append(entry)
    else:
        # one or more categories specified
        category_list = module.params["category"]

    for category in category_list:
        command_list = []
        # Build Command list for each Category
        if category in CATEGORY_COMMANDS_ALL:
            if not module.params["command"]:
                # True if we don't specify a command --> use default
                command_list.append(CATEGORY_COMMANDS_DEFAULT[category])
            elif "all" in module.params["command"]:
                for entry in CATEGORY_COMMANDS_ALL[category]:
                    command_list.append(entry)
            # one or more commands
            else:
                command_list = module.params["command"]
                # Verify that all commands are valid
                for cmd in command_list:
                    # Fail if even one command given is invalid
                    if cmd not in CATEGORY_COMMANDS_ALL[category]:
                        module.fail_json(msg=f"Invalid Command: {cmd}")
        else:
            # Fail if even one category given is invalid
            module.fail_json(msg=f"Invalid Category: {category}")

        # Organize by Categories / Commands
        if category == "Sessions":
            for command in command_list:
                if command == "GetiLOSessions":
                    result[command] = rf_utils.get_ilo_sessions()

    module.exit_json(ilo_redfish_info=result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import time
from functools import wraps
from typing import Any
from typing import TypedDict

from ansible.errors import AnsibleConnectionFailure
from ansible.module_utils.common.text.converters import to_text


def ssm_retry(func: Any) -> Any:
    """
    Decorator to retry in the case of a connection failure
    Will retry if:
    * an exception is caught
    Will not retry if
    * remaining_tries is <2
    * retries limit reached
    """

    @wraps(func)
    def wrapped(self, *args: Any, **kwargs: Any) -> Any:
        reconnection_retries = getattr(self, "reconnection_retries")
        remaining_tries = int(reconnection_retries) + 1
        verbosity_d = getattr(self, "verbosity_display", None)
        cmd_summary = f"{args[0]}..."
        for attempt in range(remaining_tries):
            try:
                return_tuple = func(self, *args, **kwargs)
                verbosity_d(4, f"ssm_retry: (success) {to_text(return_tuple)}")
                return return_tuple
            except (AnsibleConnectionFailure, Exception) as e:
                if attempt == remaining_tries - 1:
                    raise
                pause = 2**attempt - 1
                pause = min(pause, 30)

                if isinstance(e, AnsibleConnectionFailure):
                    msg = f"ssm_retry: attempt: {attempt}, cmd ({cmd_summary}), pausing for {pause} seconds"
                else:
                    msg = (
                        f"ssm_retry: attempt: {attempt}, caught exception({e})"
                        f"from cmd ({cmd_summary}),pausing for {pause} seconds"
                    )

                verbosity_d(2, msg)

                time.sleep(pause)

                # Do not attempt to reuse the existing session on retries
                # This will cause the SSM session to be completely restarted,
                # as well as reinitializing the boto3 clients
                if hasattr(self, "close"):
                    self.close()

                continue

    return wrapped


class CommandResult(TypedDict):
    """
    A dictionary that contains the executed command results.
    """

    returncode: int
    stdout: str
    stderr: str



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage Check Point Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community",
}

DOCUMENTATION = """
---
module: cp_mgmt_service_icmp6
short_description: Manages service-icmp6 objects on Check Point over Web Services API
description:
  - Manages service-icmp6 objects on Check Point devices including creating, updating and removing objects.
  - All operations are performed over Web Services API.
  - Available from R80.10 management version.
version_added: "1.0.0"
author: "Or Soffer (@chkp-orso)"
options:
  name:
    description:
      - Object name.
    type: str
    required: True
  icmp_code:
    description:
      - As listed in, <a href="http,//www.iana.org/assignments/icmp-parameters" target="_blank">RFC 792</a>.
    type: int
  icmp_type:
    description:
      - As listed in, <a href="http,//www.iana.org/assignments/icmp-parameters" target="_blank">RFC 792</a>.
    type: int
  keep_connections_open_after_policy_installation:
    description:
      - Keep connections open after policy has been installed even if they are not allowed under the new policy. This overrides the settings in the
        Connection Persistence page. If you change this property, the change will not affect open connections, but only future connections.
    type: bool
  tags:
    description:
      - Collection of tag identifiers.
    type: list
    elements: str
  color:
    description:
      - Color of the object. Should be one of existing colors.
    type: str
    choices: ['aquamarine', 'black', 'blue', 'crete blue', 'burlywood', 'cyan', 'dark green', 'khaki', 'orchid', 'dark orange', 'dark sea green',
             'pink', 'turquoise', 'dark blue', 'firebrick', 'brown', 'forest green', 'gold', 'dark gold', 'gray', 'dark gray', 'light green', 'lemon chiffon',
             'coral', 'sea green', 'sky blue', 'magenta', 'purple', 'slate blue', 'violet red', 'navy blue', 'olive', 'orange', 'red', 'sienna', 'yellow']
  comments:
    description:
      - Comments string.
    type: str
  details_level:
    description:
      - The level of detail for some of the fields in the response can vary from showing only the UID value of the object to a fully detailed
        representation of the object.
    type: str
    choices: ['uid', 'standard', 'full']
  groups:
    description:
      - Collection of group identifiers.
    type: list
    elements: str
  ignore_warnings:
    description:
      - Apply changes ignoring warnings.
    type: bool
  ignore_errors:
    description:
      - Apply changes ignoring errors. You won't be able to publish such a changes. If ignore-warnings flag was omitted - warnings will also be ignored.
    type: bool
extends_documentation_fragment: check_point.mgmt.checkpoint_objects
"""

EXAMPLES = """
- name: add-service-icmp6
  cp_mgmt_service_icmp6:
    icmp_code: 7
    icmp_type: 5
    name: Icmp1
    state: present

- name: set-service-icmp6
  cp_mgmt_service_icmp6:
    icmp_code: 13
    icmp_type: 45
    name: icmp1
    state: present

- name: delete-service-icmp6
  cp_mgmt_service_icmp6:
    name: icmp2
    state: absent
"""

RETURN = """
cp_mgmt_service_icmp6:
  description: The checkpoint object created or updated.
  returned: always, except when deleting the object.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import (
    checkpoint_argument_spec_for_objects,
    api_call,
)


def main():
    argument_spec = dict(
        name=dict(type="str", required=True),
        icmp_code=dict(type="int"),
        icmp_type=dict(type="int"),
        keep_connections_open_after_policy_installation=dict(type="bool"),
        tags=dict(type="list", elements="str"),
        color=dict(
            type="str",
            choices=[
                "aquamarine",
                "black",
                "blue",
                "crete blue",
                "burlywood",
                "cyan",
                "dark green",
                "khaki",
                "orchid",
                "dark orange",
                "dark sea green",
                "pink",
                "turquoise",
                "dark blue",
                "firebrick",
                "brown",
                "forest green",
                "gold",
                "dark gold",
                "gray",
                "dark gray",
                "light green",
                "lemon chiffon",
                "coral",
                "sea green",
                "sky blue",
                "magenta",
                "purple",
                "slate blue",
                "violet red",
                "navy blue",
                "olive",
                "orange",
                "red",
                "sienna",
                "yellow",
            ],
        ),
        comments=dict(type="str"),
        details_level=dict(type="str", choices=["uid", "standard", "full"]),
        groups=dict(type="list", elements="str"),
        ignore_warnings=dict(type="bool"),
        ignore_errors=dict(type="bool"),
    )
    argument_spec.update(checkpoint_argument_spec_for_objects)

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True
    )
    api_call_object = "service-icmp6"

    result = api_call(module, api_call_object)
    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import traceback
import typing as t
from collections import OrderedDict, defaultdict

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.dns.plugins.module_utils._argspec import (
    ArgumentSpec,
    ModuleOptionProvider,
)
from ansible_collections.community.dns.plugins.module_utils._conversion.base import (
    DNSConversionError,
)
from ansible_collections.community.dns.plugins.module_utils._conversion.converter import (
    RecordConverter,
)
from ansible_collections.community.dns.plugins.module_utils._options import (
    create_bulk_operations_argspec,
    create_record_transformation_argspec,
)
from ansible_collections.community.dns.plugins.module_utils._record import (
    DNSRecord,
    format_records_for_output,
    format_ttl,
)
from ansible_collections.community.dns.plugins.module_utils._record_set import (
    DNSRecordSet,
    format_record_set_for_output,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    DNSAPIAuthenticationError,
    DNSAPIError,
    ZoneRecordAPI,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_helpers import (
    bulk_apply_changes,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_set_helpers import (
    bulk_apply_changes as rrset_bulk_apply_changes,
)

from ._utils import get_prefix, normalize_dns_name

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from .._provider import ProviderInformation  # pragma: no cover
    from .._record import RecordIDT  # pragma: no cover
    from .._record_set import RecordSetIDT  # pragma: no cover
    from .._zone import ZoneIDT  # pragma: no cover
    from .._zone_record_set_api import ZoneRecordSetAPI  # pragma: no cover

    class RecordSetDict(t.TypedDict):  # pragma: no cover
        record: str  # pragma: no cover  # input might be None
        prefix: str | None  # pragma: no cover
        type: str  # pragma: no cover
        ttl: int | None  # pragma: no cover
        value: list[str] | None  # pragma: no cover  # never None if ignore == False
        ignore: bool  # pragma: no cover


def create_module_argument_spec(
    provider_information: ProviderInformation,
) -> ArgumentSpec:
    return (
        ArgumentSpec(
            argument_spec={
                "zone_name": {"type": "str", "aliases": ["zone"]},
                "zone_id": {"type": provider_information.get_zone_id_type()},
                "prune": {"type": "bool", "default": False},
                "record_sets": {
                    "type": "list",
                    "elements": "dict",
                    "required": True,
                    "aliases": ["records"],
                    "options": {
                        "record": {"type": "str"},
                        "prefix": {"type": "str"},
                        "ttl": {
                            "type": "int",
                            "default": provider_information.get_record_default_ttl(),
                        },
                        "type": {
                            "choices": provider_information.get_supported_record_types(),
                            "required": True,
                        },
                        "value": {"type": "list", "elements": "str"},
                        "ignore": {"type": "bool", "default": False},
                    },
                    "required_if": [("ignore", False, ["value"])],
                    "required_one_of": [("record", "prefix")],
                    "mutually_exclusive": [("record", "prefix")],
                },
            },
            required_one_of=[
                ("zone_name", "zone_id"),
            ],
            mutually_exclusive=[
                ("zone_name", "zone_id"),
            ],
        )
        .merge(create_bulk_operations_argspec(provider_information))
        .merge(create_record_transformation_argspec())
    )


def _get_record_sets_dict(
    module: AnsibleModule,
    provider_information: ProviderInformation,
    record_converter: RecordConverter,
    zone_in: str,
) -> OrderedDict[tuple[str | None, str], RecordSetDict]:
    record_sets = module.params["record_sets"]
    record_sets_dict: OrderedDict[
        tuple[str | None, str],
        tuple[int, RecordSetDict],
    ] = OrderedDict()
    for index, record_set in enumerate(record_sets):
        record_set = record_set.copy()
        record_name = record_set["record"]
        prefix = record_set["prefix"]
        record_name, prefix = get_prefix(
            normalized_zone=zone_in,
            normalized_record=record_name,
            prefix=prefix,
            provider_information=provider_information,
        )
        record_set["record"] = record_name
        record_set["prefix"] = prefix
        if record_set["value"]:
            record_set["value"] = record_converter.process_values_from_user(
                record_set["type"], record_set["value"]
            )
        if record_set["type"] not in provider_information.get_supported_record_types():
            module.fail_json(
                msg=f"Found invalid record type {record_set['type']} at index #{index}"
            )
        key = (prefix, record_set["type"])
        if key in record_sets_dict:
            module.fail_json(
                msg=f"Found multiple sets for record {record_name} and type {record_set['type']}: index #{record_sets_dict[key][0]} and #{index}"
            )
        record_sets_dict[key] = (index, record_set)
    result = OrderedDict()
    for k, (dummy, v) in record_sets_dict.items():
        result[k] = v
    return result


def _run_module_record_api(
    option_provider,
    module: AnsibleModule,
    provider_information: ProviderInformation,
    record_converter: RecordConverter,
    api: ZoneRecordAPI[ZoneIDT, RecordIDT],
) -> t.NoReturn:
    # Get zone information
    if module.params["zone_name"] is not None:
        zone_in = normalize_dns_name(module.params["zone_name"])
        zone = api.get_zone_with_records_by_name(zone_in)
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_id = zone.zone.id
        zone_records = zone.records
    else:
        zone = api.get_zone_with_records_by_id(module.params["zone_id"])
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_in = normalize_dns_name(zone.zone.name)
        zone_id = zone.zone.id
        zone_records = zone.records

    record_converter.process_multiple_from_api(zone_records)

    # Process parameters
    prune = module.params["prune"]
    record_sets_dict = _get_record_sets_dict(
        module, provider_information, record_converter, zone_in
    )

    # Group existing record sets
    existing_record_sets: dict[tuple[str | None, str], list[DNSRecord]] = {}
    for record in zone_records:
        key = (record.prefix, record.type)
        if key not in existing_record_sets:
            existing_record_sets[key] = []
        existing_record_sets[key].append(record)

    # Data required for diff
    old_record_sets = {
        k: [r.clone() for r in v] for k, v in existing_record_sets.items()
    }
    new_record_sets = {k: list(v) for k, v in existing_record_sets.items()}

    # Create action lists
    to_create = []
    to_delete = []
    to_change = []
    for (prefix, record_type), record_set_dict in record_sets_dict.items():
        key = (prefix, record_type)
        if key not in new_record_sets:
            new_record_sets[key] = []
        existing_recs = existing_record_sets.get(key, [])
        existing_record_sets[key] = []
        new_recs = new_record_sets[key]

        if record_set_dict["ignore"]:
            continue
        assert record_set_dict["value"] is not None  # since ignore==False

        mismatch_recs = []
        keep_record_sets = []
        values = list(record_set_dict["value"])
        for record in existing_recs:
            if record.ttl != record_set_dict["ttl"]:
                mismatch_recs.append(record)
                new_recs.remove(record)
                continue
            if record.target in values:
                values.remove(record.target)
                keep_record_sets.append(record)
            else:
                mismatch_recs.append(record)
                new_recs.remove(record)

        for target in values:
            if mismatch_recs:
                record = mismatch_recs.pop()
                to_change.append(record)
                new_record: DNSRecord[RecordIDT | None] = record  # type: ignore
            else:
                # Otherwise create new record
                new_record = DNSRecord(
                    record_id=None, record_type=record_type, target=target
                )
                to_create.append(new_record)
            new_record.prefix = prefix
            new_record.type = record_type
            new_record.ttl = record_set_dict["ttl"]
            new_record.target = target
            new_recs.append(new_record)

        to_delete.extend(mismatch_recs)

    # If pruning, remove superfluous record sets
    if prune:
        for key, record_set in existing_record_sets.items():
            to_delete.extend(record_set)
            for record in record_set:
                new_record_sets[key].remove(record)

    # Compose result
    result = {
        "changed": False,
        "zone_id": zone_id,
    }

    # Apply changes
    if to_create or to_delete or to_change:
        records_to_delete = record_converter.clone_multiple_to_api(to_delete)
        records_to_change = record_converter.clone_multiple_to_api(to_change)
        records_to_create = record_converter.clone_multiple_to_api(to_create)
        result["changed"] = True
        if not module.check_mode:
            dummy, errors, dummy2 = bulk_apply_changes(
                api,
                zone_id=zone_id,
                records_to_delete=records_to_delete,
                records_to_change=records_to_change,
                records_to_create=records_to_create,
                provider_information=provider_information,
                options=option_provider,
            )
            if errors:
                if len(errors) == 1:
                    raise errors[0]
                module.fail_json(
                    msg=f"Errors: {'; '.join([str(e) for e in errors])}",
                    errors=[str(e) for e in errors],
                )

    # Include diff information
    if module._diff:

        def sort_items(dictionary):
            items = [
                (
                    zone_in if prefix is None else (prefix + "." + zone_in),
                    record_type,
                    prefix,
                    record_set,
                )
                for (prefix, record_type), record_set in dictionary.items()
                if len(record_set) > 0
            ]
            return sorted(items)

        result["diff"] = {
            "before": {
                "record_sets": [
                    format_records_for_output(
                        record_set,
                        record_name,
                        prefix,
                        record_converter=record_converter,
                    )
                    for record_name, record_type, prefix, record_set in sort_items(
                        old_record_sets
                    )
                ],
            },
            "after": {
                "record_sets": [
                    format_records_for_output(
                        record_set,
                        record_name,
                        prefix,
                        record_converter=record_converter,
                    )
                    for record_name, record_type, prefix, record_set in sort_items(
                        new_record_sets
                    )
                ],
            },
        }

    module.exit_json(**result)


def _run_module_record_set_api(
    option_provider,
    module: AnsibleModule,
    provider_information: ProviderInformation,
    record_converter: RecordConverter,
    api: ZoneRecordSetAPI[ZoneIDT, RecordSetIDT, RecordIDT],
) -> t.NoReturn:
    # Get zone information
    if module.params["zone_name"] is not None:
        zone_in = normalize_dns_name(module.params["zone_name"])
        zone = api.get_zone_with_record_sets_by_name(zone_in)
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_id = zone.zone.id
        zone_record_sets = zone.record_sets
    else:
        zone = api.get_zone_with_record_sets_by_id(module.params["zone_id"])
        if zone is None:
            module.fail_json(msg="Zone not found")
        zone_in = normalize_dns_name(zone.zone.name)
        zone_id = zone.zone.id
        zone_record_sets = zone.record_sets

    for record_set in zone_record_sets:
        record_converter.process_set_from_api(record_set)

    # Process parameters
    prune = module.params["prune"]
    record_sets_dict = _get_record_sets_dict(
        module, provider_information, record_converter, zone_in
    )

    # Group existing record sets
    existing_record_sets: OrderedDict[
        tuple[str | None, str], DNSRecordSet[RecordSetIDT, RecordIDT]
    ] = OrderedDict()
    for record_set in zone_record_sets:
        key = (record_set.prefix, record_set.type)
        existing_record_sets[key] = record_set

    # Data required for diff
    old_record_sets = {k: v.clone() for k, v in existing_record_sets.items()}
    new_record_sets: dict[
        tuple[str | None, str],
        DNSRecordSet[RecordSetIDT | None, RecordIDT | None]
        | DNSRecordSet[RecordSetIDT, RecordIDT],
    ] = dict(existing_record_sets)

    # Create action lists
    to_create: list[DNSRecordSet[RecordSetIDT | None, RecordIDT | None]] = []
    to_delete: list[DNSRecordSet[RecordSetIDT, RecordIDT]] = []
    to_change: list[tuple[DNSRecordSet[RecordSetIDT, RecordIDT], bool, bool]] = []
    for (prefix, record_type), record_set_dict in record_sets_dict.items():
        key = (prefix, record_type)
        new_rrset = existing_record_sets.pop(key, None)
        if record_set_dict["ignore"]:
            continue
        assert record_set_dict["value"] is not None  # since ignore==False

        ttl = record_set_dict["ttl"]
        values = sorted(record_set_dict["value"])
        if not values:
            if new_rrset:
                to_delete.append(new_rrset)
                new_record_sets.pop(key)
            continue

        if new_rrset is None:
            rrset: DNSRecordSet[RecordSetIDT | None, RecordIDT | None] = DNSRecordSet(
                record_set_id=None, record_type=record_type
            )
            rrset.prefix = prefix
            rrset.ttl = ttl
            rrset.records = []
            for value in values:
                rec: DNSRecord[RecordIDT | None] = DNSRecord(
                    record_id=None, record_type=record_type, target=value
                )
                rec.prefix = prefix
                rec.ttl = ttl
                rrset.records.append(rec)
            to_create.append(rrset)
            new_record_sets[key] = rrset
            continue

        mismatch_ttl = ttl != new_rrset.ttl
        mismatch_values = values != sorted(rec.target for rec in new_rrset.records)
        if not mismatch_ttl and not mismatch_values:
            continue

        to_change.append((new_rrset, mismatch_values, mismatch_ttl))
        existing_records = defaultdict(list)
        for rec in new_rrset.records:
            existing_records[rec.target].append(rec)

        new_rrset.records[:] = []
        new_rrset.ttl = ttl
        for value in values:
            recs = existing_records.get(value)
            if recs:
                new_rrset.records.append(recs.pop())
                continue
            rec = DNSRecord(record_id=None, record_type=record_type, target=value)
            rec.prefix = prefix
            new_rrset.records.append(rec)

    # If pruning, remove superfluous record sets
    if prune:
        for key, record_set in existing_record_sets.items():
            to_delete.append(record_set)
            new_record_sets.pop(key)

    # Compose result
    result = {
        "changed": False,
        "zone_id": zone_id,
    }

    def _get_name(prefix):
        return zone_in if prefix is None else (prefix + "." + zone_in)

    # Apply changes
    if to_create or to_delete or to_change:
        record_sets_to_delete = [
            record_converter.clone_set_to_api(rrset) for rrset in to_delete
        ]
        record_sets_to_change = [
            (record_converter.clone_set_to_api(rrset), mismatch_values, mismatch_ttl)
            for rrset, mismatch_values, mismatch_ttl in to_change
        ]
        record_sets_to_create = [
            record_converter.clone_set_to_api(rrset) for rrset in to_create
        ]
        result["changed"] = True
        if not module.check_mode:
            dummy, errors, dummy2 = rrset_bulk_apply_changes(
                api,
                zone_id=zone_id,
                record_sets_to_delete=record_sets_to_delete,
                record_sets_to_change=record_sets_to_change,
                record_sets_to_create=record_sets_to_create,
                provider_information=provider_information,
                options=option_provider,
            )
            if errors:
                messages = [
                    (
                        f"{what.title()} record set {record_set.type} {_get_name(record_set.prefix)} with TTL={format_ttl(record_set.ttl)} and"
                        f" value={record_converter.process_values_to_user(record_set.type, [rec.target for rec in record_set.records])}: {e}"
                    )
                    for what, record_set, e in errors
                ]
                module.fail_json(
                    msg=f"Errors: {'; '.join(messages)}",
                    errors=[str(e) for dummy, dummy2, e in errors],
                )

    # Include diff information
    if module._diff:

        def sort_items(dictionary):
            items = [
                (_get_name(prefix), record_type, prefix, record_set)
                for (prefix, record_type), record_set in dictionary.items()
            ]
            return sorted(items)

        result["diff"] = {
            "before": {
                "record_sets": [
                    format_record_set_for_output(
                        record_set,
                        record_name,
                        prefix,
                        record_converter=record_converter,
                    )
                    for record_name, record_type, prefix, record_set in sort_items(
                        old_record_sets
                    )
                ],
            },
            "after": {
                "record_sets": [
                    format_record_set_for_output(
                        record_set,
                        record_name,
                        prefix,
                        record_converter=record_converter,
                    )
                    for record_name, record_type, prefix, record_set in sort_items(
                        new_record_sets
                    )
                ],
            },
        }

    module.exit_json(**result)


def run_module(
    module: AnsibleModule,
    create_api: t.Callable[[], ZoneRecordAPI | ZoneRecordSetAPI],
    provider_information: ProviderInformation,
) -> t.NoReturn:
    option_provider = ModuleOptionProvider(module)
    record_converter = RecordConverter(provider_information, option_provider)
    record_converter.emit_deprecations(module.deprecate)

    try:
        # Create API
        api = create_api()

        if isinstance(api, ZoneRecordAPI):
            _run_module_record_api(
                option_provider, module, provider_information, record_converter, api
            )
        else:
            _run_module_record_set_api(
                option_provider, module, provider_information, record_converter, api
            )

    except DNSConversionError as e:
        module.fail_json(
            msg=f"Error while converting DNS values: {e.error_message}",
            error=e.error_message,
            exception=traceback.format_exc(),
        )
    except DNSAPIAuthenticationError as e:
        module.fail_json(
            msg=f"Cannot authenticate: {e}",
            error=to_text(e),
            exception=traceback.format_exc(),
        )
    except DNSAPIError as e:
        module.fail_json(
            msg=f"Error: {e}",
            error=to_text(e),
            exception=traceback.format_exc(),
        )



#!/usr/bin/python

# Copyright (C) 2018 IBM CORPORATION
# Author(s): Tzur Eliyahu <tzure@il.ibm.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ibm_sa_host_ports
short_description: Add host ports on IBM Spectrum Accelerate Family storage systems

description:
  - This module adds ports to or removes them from the hosts on IBM Spectrum Accelerate Family storage systems.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  host:
    description:
      - Host name.
    required: true
    type: str
  state:
    description:
      - Host ports state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  iscsi_name:
    description:
      - The iSCSI initiator name.
    type: str
  fcaddress:
    description:
      - Fiber channel address.
    type: str
  num_of_visible_targets:
    description:
      - Number of visible targets.
    type: str

extends_documentation_fragment:
  - community.general._ibm_storage
  - community.general._attributes

author:
  - Tzur Eliyahu (@tzure)
"""

EXAMPLES = r"""
- name: Add ports for host.
  community.general.ibm_sa_host_ports:
    host: test_host
    iscsi_name: iqn.1994-05.com***
    username: admin
    password: secret
    endpoints: hostdev-system
    state: present

- name: Remove ports for host.
  community.general.ibm_sa_host_ports:
    host: test_host
    iscsi_name: iqn.1994-05.com***
    username: admin
    password: secret
    endpoints: hostdev-system
    state: absent
"""
RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ibm_sa_utils import (
    connect_ssl,
    execute_pyxcli_command,
    is_pyxcli_installed,
    spectrum_accelerate_spec,
)


def main():
    argument_spec = spectrum_accelerate_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["present", "absent"]),
            host=dict(required=True),
            iscsi_name=dict(),
            fcaddress=dict(),
            num_of_visible_targets=dict(),
        )
    )

    module = AnsibleModule(argument_spec)
    is_pyxcli_installed(module)

    xcli_client = connect_ssl(module)
    # required args
    ports = []
    try:
        ports = xcli_client.cmd.host_list_ports(host=module.params.get("host")).as_list
    except Exception:
        pass
    state = module.params["state"]
    port_exists = False
    ports = [port.get("port_name") for port in ports]

    fc_ports = module.params.get("fcaddress") if module.params.get("fcaddress") else []
    iscsi_ports = module.params.get("iscsi_name") if module.params.get("iscsi_name") else []
    for port in ports:
        if port in iscsi_ports or port in fc_ports:
            port_exists = True
            break
    state_changed = False
    if state == "present" and not port_exists:
        state_changed = execute_pyxcli_command(module, "host_add_port", xcli_client)
    if state == "absent" and port_exists:
        state_changed = execute_pyxcli_command(module, "host_remove_port", xcli_client)

    module.exit_json(changed=state_changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_realm_info

short_description: Allows obtaining Keycloak realm public information using Keycloak API

version_added: 4.3.0

description:
  - This module allows you to get Keycloak realm public information using the Keycloak REST API.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/8.0/rest-api/index.html).
  - Attributes are multi-valued in the Keycloak API. All attributes are lists of individual values and are returned that way
    by this module. You may pass single values for attributes when calling the module, and this is translated into a list
    suitable for the API.
extends_documentation_fragment:
  - community.general._attributes
  - community.general._attributes.info_module

options:
  auth_keycloak_url:
    description:
      - URL to the Keycloak instance.
    type: str
    required: true
    aliases:
      - url
  validate_certs:
    description:
      - Verify TLS certificates (do not disable this in production).
    type: bool
    default: true

  realm:
    type: str
    description:
      - They Keycloak realm ID.
    default: 'master'

author:
  - Fynn Chen (@fynncfchen)
"""

EXAMPLES = r"""
- name: Get a Keycloak public key
  community.general.keycloak_realm_info:
    realm: MyCustomRealm
    auth_keycloak_url: https://auth.example.com/auth
  delegate_to: localhost
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str

realm_info:
  description:
    - Representation of the realm public information.
  returned: always
  type: dict
  contains:
    realm:
      description: Realm ID.
      type: str
      returned: always
      sample: MyRealm
    public_key:
      description: Public key of the realm.
      type: str
      returned: always
      sample: MIIBIjANBgkqhkiG9w0BAQEFAAO...
    token-service:
      description: Token endpoint URL.
      type: str
      returned: always
      sample: https://auth.example.com/auth/realms/MyRealm/protocol/openid-connect
    account-service:
      description: Account console URL.
      type: str
      returned: always
      sample: https://auth.example.com/auth/realms/MyRealm/account
    tokens-not-before:
      description: The token not before.
      type: int
      returned: always
      sample: 0
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import KeycloakAPI


def main():
    """
    Module execution

    :return:
    """
    argument_spec = dict(
        auth_keycloak_url=dict(type="str", aliases=["url"], required=True, no_log=False),
        validate_certs=dict(type="bool", default=True),
        realm=dict(default="master"),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    result = dict(changed=False, msg="", realm_info="")

    kc = KeycloakAPI(module, {})

    realm = module.params.get("realm")

    realm_info = kc.get_realm_info_by_id(realm=realm)

    result["realm_info"] = realm_info
    result["msg"] = f"Get realm public info successful for ID {realm}"
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_config_templates
short_description: Resource module for organizations _config _templates
description:
  - Manage operations create, update and delete of the resource organizations _config
    _templates.
  - Create a new configuration template.
  - Remove a configuration template.
  - Update a configuration template.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  configTemplateId:
    description: ConfigTemplateId path parameter. Config template ID.
    type: str
  copyFromNetworkId:
    description: The ID of the network or config template to copy configuration from.
    type: str
  name:
    description: The name of the configuration template.
    type: str
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  timeZone:
    description: The timezone of the configuration template. For a list of allowed
      timezones, please see the 'TZ' column in the table in <a target='_blank' href='https
      //en.wikipedia.org/wiki/List_of_tz_... article</a>. Not applicable if copying
      from existing network or template.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationConfigTemplate
    description: Complete reference of the createOrganizationConfigTemplate API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-config-template
  - name: Cisco Meraki documentation for organizations deleteOrganizationConfigTemplate
    description: Complete reference of the deleteOrganizationConfigTemplate API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-config-template
  - name: Cisco Meraki documentation for organizations updateOrganizationConfigTemplate
    description: Complete reference of the updateOrganizationConfigTemplate API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-config-template
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_config_template,
    organizations.Organizations.delete_organization_config_template,
    organizations.Organizations.update_organization_config_template,
  - Paths used are
    post /organizations/{organizationId}/configTemplates,
    delete /organizations/{organizationId}/configTemplates/{configTemplateId},
    put /organizations/{organizationId}/configTemplates/{configTemplateId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_config_templates:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    copyFromNetworkId: N_24329156
    name: My config template
    organizationId: string
    timeZone: America/Los_Angeles
- name: Update by id
  cisco.meraki.organizations_config_templates:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    configTemplateId: string
    name: My config template
    organizationId: string
    timeZone: America/Los_Angeles
- name: Delete by id
  cisco.meraki.organizations_config_templates:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    configTemplateId: string
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "name": "string",
      "productTypes": [
        "string"
      ],
      "timeZone": "string"
    }
"""



#!/usr/bin/python

# Copyright (c) 2022, Dušan Marković (@bratwurzt)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_user_rolemapping

short_description: Allows administration of Keycloak user_rolemapping with the Keycloak API

version_added: 5.7.0

description:
  - This module allows you to add, remove or modify Keycloak user_rolemapping with the Keycloak REST API. It requires access
    to the REST API using OpenID Connect; the user connecting and the client being used must have the requisite access rights.
    In a default Keycloak installation, admin-cli and an admin user would work, as would a separate client definition with
    the scope tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/8.0/rest-api/index.html).
  - Attributes are multi-valued in the Keycloak API. All attributes are lists of individual values and are returned that way
    by this module. You may pass single values for attributes when calling the module, and this is translated into a list
    suitable for the API.
  - When updating a user_rolemapping, where possible provide the role ID to the module. This removes a lookup to the API to
    translate the name into the role ID.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the user_rolemapping.
      - On V(present), the user_rolemapping is created if it does not yet exist, or updated with the parameters you provide.
      - On V(absent), the user_rolemapping is removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  realm:
    type: str
    description:
      - They Keycloak realm under which this role_representation resides.
    default: 'master'

  target_username:
    type: str
    description:
      - Username of the user roles are mapped to.
      - This parameter is not required (can be replaced by uid for less API call).
  uid:
    type: str
    description:
      - ID of the user to be mapped.
      - This parameter is not required for updating or deleting the rolemapping but providing it reduces the number of API
        calls required.
  service_account_user_client_id:
    type: str
    description:
      - Client ID of the service-account-user to be mapped.
      - This parameter is not required for updating or deleting the rolemapping but providing it reduces the number of API
        calls required.
  client_id:
    type: str
    description:
      - Name of the client (different than O(cid)) whose role is to be mapped.
      - This parameter is required if O(cid) is not provided (can be replaced by O(cid) to reduce the number of API calls
        that must be made).
      - If neither O(cid) nor O(client_id) is specified, a B(realm) role is mapped instead.
  cid:
    type: str
    description:
      - ID of the client whose role is to be mapped.
      - This parameter is not required for updating or deleting the rolemapping but providing it reduces the number of API
        calls required.
      - If neither O(cid) nor O(client_id) is specified, a B(realm) role is mapped instead.
  roles:
    description:
      - Roles to be mapped to the user.
    type: list
    elements: dict
    suboptions:
      name:
        type: str
        description:
          - Name of the role representation.
          - This parameter is required only when creating or updating the role_representation.
      id:
        type: str
        description:
          - The unique identifier for this role_representation.
          - This parameter is not required for updating or deleting a role_representation but providing it reduces the number
            of API calls required.
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Dušan Marković (@bratwurzt)
  - Ivan Kokalović (@koke1997)
"""

EXAMPLES = r"""
- name: Map a realm role to a user, authentication with credentials
  community.general.keycloak_user_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: present
    uid: user_uid
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Map a client role to a user, authentication with credentials
  community.general.keycloak_user_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: present
    client_id: client1
    uid: user_uid
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Map a client role to a service account user for a client, authentication with credentials
  community.general.keycloak_user_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: present
    client_id: client1
    service_account_user_client_id: clientIdOfServiceAccount
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Map a client role to a user, authentication with token
  community.general.keycloak_user_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
    state: present
    client_id: client1
    target_username: user1
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Unmap client role from a user
  community.general.keycloak_user_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: absent
    client_id: client1
    uid: 70e3ae72-96b6-11e6-9056-9737fd4d0764
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "Role role1 assigned to user user1."

proposed:
  description: Representation of proposed client role mapping.
  returned: always
  type: dict
  sample: {"clientId": "test"}

existing:
  description:
    - Representation of existing client role mapping.
    - The sample is truncated.
  returned: always
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }

end_state:
  description:
    - Representation of client role mapping after module execution.
    - The sample is truncated.
  returned: on success
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    roles_spec = dict(
        name=dict(type="str"),
        id=dict(type="str"),
    )

    meta_args = dict(
        state=dict(default="present", choices=["present", "absent"]),
        realm=dict(default="master"),
        uid=dict(type="str"),
        target_username=dict(type="str"),
        service_account_user_client_id=dict(type="str"),
        cid=dict(type="str"),
        client_id=dict(type="str"),
        roles=dict(type="list", elements="dict", options=roles_spec),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [
                ["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"],
                ["uid", "target_username", "service_account_user_client_id"],
            ]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", diff={}, proposed={}, existing={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    state = module.params.get("state")
    cid = module.params.get("cid")
    client_id = module.params.get("client_id")
    uid = module.params.get("uid")
    target_username = module.params.get("target_username")
    service_account_user_client_id = module.params.get("service_account_user_client_id")
    roles = module.params.get("roles")

    # Check the parameters
    if uid is None and target_username is None and service_account_user_client_id is None:
        module.fail_json(
            msg="Either the `target_username`, `uid` or `service_account_user_client_id` has to be specified."
        )

    # Get the potential missing parameters
    if uid is None and service_account_user_client_id is None:
        user_rep = kc.get_user_by_username(username=target_username, realm=realm)
        if user_rep is not None:
            uid = user_rep.get("id")
        else:
            module.fail_json(msg=f"Could not fetch user for username {target_username}:")
    else:
        if uid is None and target_username is None:
            user_rep = kc.get_service_account_user_by_client_id(client_id=service_account_user_client_id, realm=realm)
            if user_rep is not None:
                uid = user_rep["id"]
            else:
                module.fail_json(msg=f"Could not fetch service-account-user for client_id {target_username}:")

    if cid is None and client_id is not None:
        cid = kc.get_client_id(client_id=client_id, realm=realm)
        if cid is None:
            module.fail_json(msg=f"Could not fetch client {client_id}:")
    if roles is None:
        module.exit_json(msg="Nothing to do (no roles specified).")
    else:
        for role in roles:
            if role.get("name") is None and role.get("id") is None:
                module.fail_json(msg="Either the `name` or `id` has to be specified on each role.")
            # Fetch missing role_id
            if role.get("id") is None:
                if cid is None:
                    role_id = kc.get_realm_role(name=role.get("name"), realm=realm)["id"]
                else:
                    role_id = kc.get_client_role_id_by_name(cid=cid, name=role.get("name"), realm=realm)
                if role_id is not None:
                    role["id"] = role_id
                else:
                    module.fail_json(
                        msg=f"Could not fetch role {role.get('name')} for client_id {client_id} or realm {realm}"
                    )
            # Fetch missing role_name
            else:
                if cid is None:
                    role_rep = kc.get_realm_user_rolemapping_by_id(uid=uid, rid=role.get("id"), realm=realm)
                    if role_rep is not None:
                        role["name"] = role_rep["name"]
                else:
                    role_rep = kc.get_client_user_rolemapping_by_id(uid=uid, cid=cid, rid=role.get("id"), realm=realm)
                    if role_rep is not None:
                        role["name"] = role_rep["name"]
                if role.get("name") is None:
                    module.fail_json(
                        msg=f"Could not fetch role {role.get('id')} for client_id {client_id} or realm {realm}"
                    )

    # Get effective role mappings
    if cid is None:
        available_roles_before = kc.get_realm_user_available_rolemappings(uid=uid, realm=realm)
        assigned_roles_before = kc.get_realm_user_composite_rolemappings(uid=uid, realm=realm)
    else:
        available_roles_before = kc.get_client_user_available_rolemappings(uid=uid, cid=cid, realm=realm)
        assigned_roles_before = kc.get_client_user_composite_rolemappings(uid=uid, cid=cid, realm=realm)

    result["existing"] = assigned_roles_before
    result["proposed"] = roles

    update_roles = []
    for role in roles:
        # Fetch roles to assign if state present
        if state == "present":
            for available_role in available_roles_before:
                if role.get("name") == available_role.get("name"):
                    update_roles.append(
                        {
                            "id": role.get("id"),
                            "name": role.get("name"),
                        }
                    )
        # Fetch roles to remove if state absent
        else:
            for assigned_role in assigned_roles_before:
                if role.get("name") == assigned_role.get("name"):
                    update_roles.append(
                        {
                            "id": role.get("id"),
                            "name": role.get("name"),
                        }
                    )

    if len(update_roles):
        if state == "present":
            # Assign roles
            result["changed"] = True
            if module._diff:
                result["diff"] = dict(before={"roles": assigned_roles_before}, after={"roles": update_roles})
            if module.check_mode:
                module.exit_json(**result)
            kc.add_user_rolemapping(uid=uid, cid=cid, role_rep=update_roles, realm=realm)
            result["msg"] = f"Roles {update_roles} assigned to userId {uid}."
            if cid is None:
                assigned_roles_after = kc.get_realm_user_composite_rolemappings(uid=uid, realm=realm)
            else:
                assigned_roles_after = kc.get_client_user_composite_rolemappings(uid=uid, cid=cid, realm=realm)
            result["end_state"] = assigned_roles_after
            module.exit_json(**result)
        else:
            # Remove mapping of role
            result["changed"] = True
            if module._diff:
                result["diff"] = dict(before={"roles": assigned_roles_before}, after={"roles": update_roles})
            if module.check_mode:
                module.exit_json(**result)
            kc.delete_user_rolemapping(uid=uid, cid=cid, role_rep=update_roles, realm=realm)
            result["msg"] = f"Roles {update_roles} removed from userId {uid}."
            if cid is None:
                assigned_roles_after = kc.get_realm_user_composite_rolemappings(uid=uid, realm=realm)
            else:
                assigned_roles_after = kc.get_client_user_composite_rolemappings(uid=uid, cid=cid, realm=realm)
            result["end_state"] = assigned_roles_after
            module.exit_json(**result)
    # Do nothing
    else:
        result["changed"] = False
        result["msg"] = f"Nothing to do, roles {roles} are correctly mapped to user for username {target_username}."
        module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2021, Phillipe Smith <phsmithcc@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import traceback
import typing as t
from http import HTTPStatus

from ansible.module_utils.urls import fetch_url, url_argument_spec

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


def api_argument_spec() -> dict[str, t.Any]:
    """
    Creates an argument spec that can be used with any module
    that will be requesting content via Rundeck API
    """
    api_argument_spec = url_argument_spec()
    api_argument_spec.update(
        dict(
            url=dict(required=True, type="str"),
            api_version=dict(type="int", default=39),
            api_token=dict(required=True, type="str", no_log=True),
        )
    )

    return api_argument_spec


def api_request(
    module: AnsibleModule,
    endpoint: str,
    data: t.Any | None = None,
    method: str = "GET",
    content_type: str = "application/json",
) -> tuple[t.Any, dict[str, t.Any]]:
    """Manages Rundeck API requests via HTTP(S)

    :arg module:   The AnsibleModule (used to get url, api_version, api_token, etc).
    :arg endpoint: The API endpoint to be used.
    :kwarg data:   The data to be sent (in case of POST/PUT).
    :kwarg method: "POST", "PUT", etc.

    :returns: A tuple of (**response**, **info**). Use ``response.read()`` to read the data.
        The **info** contains the 'status' and other meta data. When a HttpError (status >= 400)
        occurred then ``info['body']`` contains the error response data::

    Example::

        data={...}
        resp, info = fetch_url(module,
                               "http://rundeck.example.org",
                               data=module.jsonify(data),
                               method="POST")
        status_code = info["status"]
        body = resp.read()
        if status_code >= 400 :
            body = info['body']
    """

    response, info = fetch_url(
        module=module,
        url=f"{module.params['url']}/api/{module.params['api_version']}/{endpoint}",
        data=json.dumps(data),
        method=method,
        headers={
            "Content-Type": content_type,
            "Accept": "application/json",
            "X-Rundeck-Auth-Token": module.params["api_token"],
        },
    )

    _status = HTTPStatus(info["status"])

    if _status == HTTPStatus.FORBIDDEN:
        module.fail_json(msg="Token authorization failed", execution_info=json.loads(info["body"]))
    elif _status == HTTPStatus.NOT_FOUND:
        return None, info
    elif _status == HTTPStatus.CONFLICT:
        module.fail_json(msg="Job executions limit reached", execution_info=json.loads(info["body"]))
    elif _status >= HTTPStatus.INTERNAL_SERVER_ERROR:  # 5xx errors
        module.fail_json(msg="Rundeck API error", execution_info=json.loads(info["body"]))

    try:
        content = response.read()

        if not content:
            return None, info
        else:
            json_response = json.loads(content)
            return json_response, info
    except AttributeError as error:
        module.fail_json(
            msg=f"Rundeck API request error: {error}", exception=traceback.format_exc(), execution_info=info
        )
    except ValueError as error:
        module.fail_json(
            msg=f"No valid JSON response: {error}", exception=traceback.format_exc(), execution_info=content
        )



#!/usr/bin/python
# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hetzner_dns_record_set_info

short_description: Retrieve record sets in Hetzner DNS service

version_added: 2.0.0

description:
  - Retrieves DNS record sets in Hetzner DNS service.
extends_documentation_fragment:
  - community.dns._hetzner
  - community.dns._hetzner.record_type_choices
  - community.dns._hetzner.record_type_seealso
  - community.dns._hetzner.zone_id_type
  - community.dns._module_record_set_info
  - community.dns._options.record_transformation
  - community.dns._attributes
  - community.dns._attributes.actiongroup_hetzner
  - community.dns._attributes.info_module
  - community.dns._attributes.idempotent_not_modify_state

attributes:
  action_group:
    version_added: 2.4.0

author:
  - Markus Bergholz (@markuman) <markuman+spambelongstogoogle@gmail.com>
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.dns.hetzner_dns_record_info
  - plugin: community.dns.hetzner_dns_records
    plugin_type: inventory
"""

EXAMPLES = r"""
- name: Retrieve the details for the A records of new.foo.com
  community.dns.hetzner_dns_record_set_info:
    zone: foo.com
    record: new.foo.com
    type: A
    hetzner_token: access_token
  register: rec

- name: Print the A record set
  ansible.builtin.debug:
    msg: "{{ rec.set }}"
"""

RETURN = r"""
set:
  description: The fetched record set. Is empty if record set does not exist.
  type: dict
  returned: success and O(what) is V(single_record)
  contains:
    record:
      description: The record name.
      type: str
      sample: sample.example.com
    prefix:
      description: The record prefix.
      type: str
      sample: sample
      version_added: 0.2.0
    type:
      description: The DNS record type.
      type: str
      sample: A
    ttl:
      description:
        - The TTL.
        - If there are records in this set with different TTLs, the minimum of the TTLs will be presented here.
        - Will return V(none) if the zone's default TTL is used.
      type: int
      sample: 3600
    ttls:
      description:
        - If there are records with different TTL values in this set, this will be the list of TTLs appearing in the records.
        - Every distinct TTL will appear once, and the TTLs are in ascending order.
      returned: When there is more than one distinct TTL
      type: list
      elements: int
      sample:
        - 300
        - 3600
    value:
      description: The DNS record set's value.
      type: list
      elements: str
      sample:
        - 1.2.3.4
        - 1.2.3.5
  sample:
    record: sample.example.com
    type: A
    ttl: 3600
    value:
      - 1.2.3.4
      - 1.2.3.5

sets:
  description: The list of fetched record sets.
  type: list
  elements: dict
  returned: success and O(what) is not V(single_record)
  contains:
    record:
      description: The record name.
      type: str
      sample: sample.example.com
    prefix:
      description: The record prefix.
      type: str
      sample: sample
      version_added: 0.2.0
    type:
      description: The DNS record type.
      type: str
      sample: A
    ttl:
      description:
        - The TTL.
        - If there are records in this set with different TTLs, the minimum of the TTLs will be presented here.
        - Will return V(none) if the zone's default TTL is used.
      type: int
      sample: 3600
    ttls:
      description:
        - If there are records with different TTL values in this set, this will be the list of TTLs appearing in the records.
        - Every distinct TTL will appear once, and the TTLs are in ascending order.
      returned: When there is more than one distinct TTL
      type: list
      elements: int
      sample:
        - 300
        - 3600
    value:
      description: The DNS record set's value.
      type: list
      elements: str
      sample:
        - 1.2.3.4
        - 1.2.3.5
  sample:
    - record: sample.example.com
      type: A
      ttl: 3600
      value:
        - 1.2.3.4
        - 1.2.3.5

zone_id:
  description: The ID of the zone.
  type: str
  returned: success
  sample: 23
  version_added: 0.2.0
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.dns.plugins.module_utils._argspec import (
    ModuleOptionProvider,
)
from ansible_collections.community.dns.plugins.module_utils._hetzner.api import (
    create_hetzner_api,
    create_hetzner_argument_spec,
    create_hetzner_provider_information,
)
from ansible_collections.community.dns.plugins.module_utils._http import (
    ModuleHTTPHelper,
)
from ansible_collections.community.dns.plugins.module_utils._module.record_set_info import (
    create_module_argument_spec,
    run_module,
)


def main() -> None:
    provider_information = create_hetzner_provider_information()
    argument_spec = create_hetzner_argument_spec()
    argument_spec.merge(
        create_module_argument_spec(provider_information=provider_information)
    )
    module = AnsibleModule(supports_check_mode=True, **argument_spec.to_kwargs())
    option_provider = ModuleOptionProvider(module)
    run_module(
        module,
        lambda: create_hetzner_api(option_provider, ModuleHTTPHelper(module)),
        provider_information=create_hetzner_provider_information(
            option_provider=option_provider
        ),
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ..retries import AWSRetry

# from .common import AnsibleAutoScalingError
from .common import AutoScalingErrorHandler


@AutoScalingErrorHandler.list_error_handler("list auto scaling instances", default_value=[])
@AWSRetry.jittered_backoff()
def describe_auto_scaling_instances(client, instance_ids=None):
    args = {}
    if instance_ids:
        args["InstanceIds"] = instance_ids

    paginator = client.get_paginator("describe_auto_scaling_instances")
    return paginator.paginate(**args).build_full_result()["AutoScalingInstances"]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Lionel Hercot (@lhercot) <lhercot@cisco.com>
# Copyright: (c) 2020, Jorge Gomez (@jgomezve) <jgomezve@cisco.com> (based on mso_dhcp_relay_policy module)
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["preview"],
    "supported_by": "community",
}

DOCUMENTATION = r"""
---
module: mso_dhcp_option_policy_option
short_description: Manage DHCP options in a DHCP Option policy.
description:
- Manage DHCP options in a DHCP Option policy on Cisco Multi-Site Orchestrator.
- This module is only supported on NDO version prior to v4.0.
author:
- Lionel Hercot (@lhercot)
options:
  dhcp_option_policy:
    description:
    - Name of the DHCP Option Policy
    type: str
    required: true
  name:
    description:
    - Name of the option in the DHCP Option Policy
    type: str
    aliases: [ option ]
  id:
    description:
    - Id of the option in the DHCP Option Policy
    type: int
  data:
    description:
    - Data of the DHCP option in the DHCP Option Policy
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new option to a DHCP Option Policy
  cisco.mso.mso_dhcp_option_policy_option:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    dhcp_option_policy: my_test_dhcp_policy
    name: ansible_test
    id: 1
    data: Data stored in the option
    state: present

- name: Remove a option to a DHCP Option Policy
  cisco.mso.mso_dhcp_option_policy_option:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    dhcp_option_policy: my_test_dhcp_policy
    name: ansible_test
    state: absent

- name: Query a option to a DHCP Option Policy
  cisco.mso.mso_dhcp_option_policy_option:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    dhcp_option_policy: my_test_dhcp_policy
    name: ansible_test
    state: query
  register: query_result

- name: Query all option of a DHCP Option Policy
  cisco.mso.mso_dhcp_option_policy_option:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    dhcp_option_policy: my_test_dhcp_policy
    state: query
  register: query_result
"""

RETURN = r"""
"""
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import (
    MSOModule,
    mso_argument_spec,
)


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        dhcp_option_policy=dict(type="str", required=True),
        name=dict(type="str", aliases=["option"]),
        id=dict(type="int"),
        data=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["name", "id", "data"]],
            ["state", "absent", ["name"]],
        ],
    )

    dhcp_option_policy = module.params.get("dhcp_option_policy")
    option_id = module.params.get("id")
    name = module.params.get("name")
    data = module.params.get("data")
    state = module.params.get("state")

    mso = MSOModule(module)

    path = "policies/dhcp/option"

    option_index = None
    previous_option = {}

    # Query for existing object(s)
    dhcp_option_obj = mso.get_obj(path, name=dhcp_option_policy, key="DhcpRelayPolicies")
    if "id" not in dhcp_option_obj:
        mso.fail_json(msg="DHCP Option Policy '{0}' is not a valid DHCP Option Policy name.".format(dhcp_option_policy))
    policy_id = dhcp_option_obj.get("id")
    options = []
    if "dhcpOption" in dhcp_option_obj:
        options = dhcp_option_obj.get("dhcpOption")
        for index, opt in enumerate(options):
            if opt.get("name") == name:
                previous_option = opt
                option_index = index

    # If we found an existing object, continue with it
    path = "{0}/{1}".format(path, policy_id)

    if state == "query":
        mso.existing = options
        if name is not None:
            mso.existing = previous_option
        mso.exit_json()

    mso.previous = previous_option
    if state == "absent":
        option = {}
        if previous_option and option_index is not None:
            options.pop(option_index)

    elif state == "present":
        option = dict(
            id=str(option_id),
            name=name,
            data=data,
        )
        if option_index is not None:
            options[option_index] = option
        else:
            options.append(option)

    if module.check_mode:
        mso.existing = option
    else:
        mso.existing = dhcp_option_obj
        dhcp_option_obj["dhcpOption"] = options
        mso.sanitize(dhcp_option_obj, collate=True)
        new_dhcp_option_obj = mso.request(path, method="PUT", data=mso.sent)
        mso.existing = {}
        for index, opt in enumerate(new_dhcp_option_obj.get("dhcpOption")):
            if opt.get("name") == name:
                mso.existing = opt

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Don Naro (@oranod) <dnaro@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = r"""
---
module: clickhouse_role

short_description: Creates or removes a ClickHouse role.

description:
  - Creates or removes a ClickHouse role.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    support: full
    description:
      - O(settings) can break idempotency when the deprecated C(list) format is used. Use the dictionary format instead.

version_added: '0.5.0'

author:
  - Don Naro (@oranod)
  - Aleksandr Vagachev (@aleksvagachev)
  - Andrew Klychkov (@Andersson007)
  - Rafal Kozlowski (@rkozlo)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts

options:
  state:
    description:
      - Role state.
      - C(present) creates the role if it does not exist.
      - C(absent) deletes the role if it exists.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  name:
    description:
      - Role name to add or remove.
    type: str
    required: true
  cluster:
    description:
      - Run the command on all cluster hosts.
      - If the cluster is not configured, the command will crash with an error.
    type: str
  settings:
    description:
      - Settings with their limitations that apply to the role.
      - Can pass either list(obsolete) or dictionary.
      - You can also specify the profile from which the settings will be inherited C(list).
    type: raw
  profiles:
    version_added: 2.2.0
    description:
      - Settings profiles that will be applied to role.
      - Can be only used with O(settings) as C(dict).
    type: list
    elements: str
    default: []
"""

EXAMPLES = r"""
- name: Create role
  community.clickhouse.clickhouse_role:
    name: test_role
    state: present

- name: Create a role with settings
  community.clickhouse.clickhouse_role:
    name: test_role
    state: present
    settings:
      max_memory_usage:
        value: 15000
        min: 15000
        max: 16000
        writability: readonly
    profiles:
      - restricted
    cluster: test_cluster

- name: Create a role with settings and profiles
  community.clickhouse.clickhouse_role:
    name: test_role
    state: present
    settings:
      max_memory_usage:
        value: 15000
        min: 15000
        max: 16000
        writability: readonly
    profiles:
      - restricted
    cluster: test_cluster

- name: Remove role
  community.clickhouse.clickhouse_role:
    name: test_role
    state: absent
"""

RETURN = r"""
executed_statements:
  description:
  - Data-modifying executed statements.
  returned: on success
  type: list
  sample: ['CREATE ROLE test_role']
"""

import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
)
from ansible_collections.community.clickhouse.plugins.module_utils.entity_settings import EntitySettings
executed_statements = []


class ClickHouseRole:
    def __init__(self, module, client, name):
        self.module = module
        self.client = client
        self.name = name
        self.settings = EntitySettings(self.module, self.client, self.name, 'role')
        self.exists = self.check_exists()

    def check_exists(self):
        query = "SELECT 1 FROM system.roles WHERE name = '%s' LIMIT 1" % self.name
        result = execute_query(self.module, self.client, query)
        return bool(result)

    def get_current_role_definition(self):
        """Get current role definition using SHOW CREATE ROLE"""
        if not self.exists:
            return None

        query = "SHOW CREATE ROLE %s" % self.name
        result = execute_query(self.module, self.client, query)
        if result:
            return result[0][0]  # SHOW CREATE ROLE returns single row with CREATE statement
        return None

    def parse_settings_from_create_statement(self, create_statement):
        """Parse settings from CREATE ROLE statement"""
        if not create_statement or 'SETTINGS' not in create_statement:
            return []

        # Extract settings part after SETTINGS keyword
        settings_part = create_statement.split('SETTINGS', 1)[1].strip()
        if not settings_part:
            return []

        # Parse individual settings (this is a simplified parser)
        # In real implementation, we might need more sophisticated parsing
        settings = []
        current_setting = ""
        paren_depth = 0
        quote_char = None

        i = 0
        while i < len(settings_part):
            char = settings_part[i]

            if quote_char:
                current_setting += char
                if char == quote_char and (i == 0 or settings_part[i - 1] != '\\'):
                    quote_char = None
            elif char in ("'", '"'):
                current_setting += char
                quote_char = char
            elif char == '(':
                current_setting += char
                paren_depth += 1
            elif char == ')':
                current_setting += char
                paren_depth -= 1
            elif char == ',' and paren_depth == 0:
                settings.append(current_setting.strip())
                current_setting = ""
            else:
                current_setting += char
            i += 1

        if current_setting.strip():
            settings.append(current_setting.strip())

        return settings

    def normalize_settings(self, settings_list):
        """Normalize settings for comparison"""
        if not settings_list:
            return []

        normalized = []
        for setting in settings_list:
            # Remove extra whitespace and standardize format
            normalized_setting = ' '.join(setting.split())

            # ClickHouse normalizes some constraint types
            # Different versions may use CONST or READONLY interchangeably
            # Since READONLY is an alias for CONST in ClickHouse, normalize both to CONST
            normalized_setting = re.sub(r'\bREADONLY\b', 'CONST', normalized_setting)

            # ClickHouse may handle profile names differently across versions
            # Normalize quoted and unquoted profile names
            # PROFILE 'default' -> PROFILE default, PROFILE "default" -> PROFILE default
            normalized_setting = re.sub(r"PROFILE\s+'([^']+)'", r"PROFILE \1", normalized_setting)
            normalized_setting = re.sub(r'PROFILE\s+"([^"]+)"', r"PROFILE \1", normalized_setting)

            # Also handle case where profile name might be output with backticks
            normalized_setting = re.sub(r"PROFILE\s+`([^`]+)`", r"PROFILE \1", normalized_setting)

            normalized.append(normalized_setting)

        return sorted(normalized)

    def settings_changed(self, current_settings, desired_settings):
        """Check if settings have changed"""
        current_normalized = self.normalize_settings(current_settings)
        desired_normalized = self.normalize_settings(desired_settings)

        # For debugging version compatibility issues
        if self.module._verbosity >= 3:  # Only show at high verbosity
            self.module.log(f"Current settings (normalized): {current_normalized}")
            self.module.log(f"Desired settings (normalized): {desired_normalized}")

        return current_normalized != desired_normalized

    def create(self):
        if not self.exists:
            query = "CREATE ROLE %s" % self.name

            if self.module.params['cluster']:
                query += " ON CLUSTER %s" % self.module.params['cluster']

            if isinstance(self.module.params['settings'], list):
                list_settings = self.module.params['settings']
                if list_settings:
                    query += " SETTINGS"
                    for index, value in enumerate(list_settings):
                        query += " %s" % value
                        if index < len(list_settings) - 1:
                            query += ","
            elif isinstance(self.module.params['settings'], dict) or self.module.params['profiles']:
                settings_entity = self.settings.compare_and_build_clause(self.module.params['settings'], self.module.params['profiles'])
                query += settings_entity[1]

            executed_statements.append(query)

            if not self.module.check_mode:
                execute_query(self.module, self.client, query)

            self.exists = True
            return True
        else:
            return False

    def alter(self, settings, profiles, cluster):
        """Update role settings using ALTER ROLE"""
        if not self.exists:
            return False

        query = "ALTER ROLE %s" % self.name

        if cluster:
            query += " ON CLUSTER %s" % cluster

        if isinstance(settings, list):
            # Handle settings updates
            if settings is not None and len(settings) > 0:
                # Set these settings
                query += " SETTINGS"
                for index, value in enumerate(settings):
                    query += " %s" % value
                    if index < len(settings) - 1:
                        query += ","
        elif isinstance(settings, dict) or (settings is None and profiles):
            settings_entity = self.settings.compare_and_build_clause(settings, profiles)
            if settings_entity[0] is False:
                return False
            query += settings_entity[1]

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        return True

    def drop(self):
        if self.exists:
            query = "DROP ROLE %s" % self.name
            executed_statements.append(query)

            if not self.module.check_mode:
                execute_query(self.module, self.client, query)

            self.exists = False
            return True
        else:
            return False


def main():
    # Set up arguments.
    # If there are common arguments shared across several modules,
    # create the common_argument_spec() function under plugins/module_utils/*
    # and invoke here to return a dict with those arguments
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        state=dict(type="str", choices=["present", "absent"], default="present"),
        name=dict(type="str", required=True),
        cluster=dict(type='str', default=None),
        settings=dict(type='raw'),
        profiles=dict(type='list', elements='str', default=[])
    )

    # Instantiate an object of module class
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    # Assign passed options to variables
    client_kwargs = module.params["client_kwargs"]
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)
    state = module.params["state"]
    name = module.params["name"]
    settings = module.params["settings"]
    profiles = module.params["profiles"]
    cluster = module.params["cluster"]

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)

    # Do the job
    changed = False
    role = ClickHouseRole(module, client, name)
    desired_settings = module.params.get("settings", [])

    if state == "present":
        if not role.exists:
            # Role doesn't exist, create it
            changed = role.create()
        else:
            # Role exists, check if settings need to be updated
            if isinstance(desired_settings, list):
                if desired_settings is not None and len(desired_settings) > 0:  # Only check settings if they are specified and not empty
                    current_definition = role.get_current_role_definition()
                    current_settings = role.parse_settings_from_create_statement(current_definition) if current_definition else []

                    if role.settings_changed(current_settings, desired_settings):
                        changed = role.alter(desired_settings, profiles, cluster)
            # No need to check settings in new settings parameter
            elif isinstance(settings, dict) or profiles:
                changed = role.alter(desired_settings, profiles, cluster)
    else:
        # If state is absent
        if role.exists:
            changed = role.drop()

    # Close connection
    client.disconnect_connection()

    # Users will get this in JSON output after execution
    module.exit_json(changed=changed, executed_statements=executed_statements)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: administered_identities_me_api_keys_revoke
short_description: Resource module for administered _identities _me _api _keys _revoke
description:
  - Manage operation create of the resource administered _identities _me _api _keys
    _revoke. - > Revokes an identity's API key, using the last four characters of
    the key. For users who have access to more than one organization, the change will
    take up to five minutes to propagate. If one of the organizations is currently
    under maintenance, the change may not propagate fully until after the maintenance
    has been completed.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  suffix:
    description: Suffix path parameter.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for administered revokeAdministeredIdentitiesMeApiKeys
    description: Complete reference of the revokeAdministeredIdentitiesMeApiKeys API.
    link: https://developer.cisco.com/meraki/api-v1/#!revoke-administered-identities-me-api-keys
notes:
  - SDK Method used are
    administered.Administered.revoke_administered_identities_me_api_keys,
  - Paths used are
    post /administered/identities/me/api/keys/{suffix}/revoke,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.administered_identities_me_api_keys_revoke:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    suffix: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {}
"""



#!/usr/bin/python
# Copyright (c) 2024 Alexander Bakanovskii <skottttt228@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipa_getkeytab
short_description: Manage keytab file in FreeIPA
version_added: 9.5.0
description:
  - Manage keytab file with C(ipa-getkeytab) utility.
  - See U(https://manpages.ubuntu.com/manpages/jammy/man1/ipa-getkeytab.1.html) for reference.
author: "Alexander Bakanovskii (@abakanovskii)"
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  path:
    description:
      - The base path where to put generated keytab file.
    type: path
    aliases: ["keytab"]
    required: true
  principal:
    description:
      - The non-realm part of the full principal name.
    type: str
    required: true
  ipa_host:
    description:
      - The IPA server to retrieve the keytab from (FQDN).
    type: str
  ldap_uri:
    description:
      - LDAP URI. If V(ldap://) is specified, STARTTLS is initiated by default.
      - Can not be used with the O(ipa_host) option.
    type: str
  bind_dn:
    description:
      - The LDAP DN to bind as when retrieving a keytab without Kerberos credentials.
      - Generally used with the O(bind_pw) option.
    type: str
  bind_pw:
    description:
      - The LDAP password to use when not binding with Kerberos.
    type: str
  password:
    description:
      - Use this password for the key instead of one randomly generated.
    type: str
  ca_cert:
    description:
      - The path to the IPA CA certificate used to validate LDAPS/STARTTLS connections.
    type: path
  sasl_mech:
    description:
      - SASL mechanism to use if O(bind_dn) and O(bind_pw) are not specified.
    choices: ["GSSAPI", "EXTERNAL"]
    type: str
  retrieve_mode:
    description:
      - Retrieve an existing key from the server instead of generating a new one.
      - This is incompatible with the O(password), and works only against a IPA server more recent than version 3.3.
      - The user requesting the keytab must have access to the keys for this operation to succeed.
      - Be aware that if set V(true), a new keytab is generated.
      - This invalidates all previously retrieved keytabs for this service principal.
    type: bool
  encryption_types:
    description:
      - The list of encryption types to use to generate keys.
      - It uses local client defaults if not provided.
      - Valid values depend on the Kerberos library version and configuration.
    type: str
  state:
    description:
      - The state of the keytab file.
      - V(present) only check for existence of a file, if you want to recreate keytab with other parameters you should set
        O(force=true).
    type: str
    default: present
    choices: ["present", "absent"]
  force:
    description:
      - Force recreation if exists already.
    type: bool
requirements:
  - freeipa-client
  - Managed host is FreeIPA client
extends_documentation_fragment:
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Get Kerberos ticket using default principal
  community.general.krb_ticket:
    password: "{{ aldpro_admin_password }}"

- name: Create keytab
  community.general.ipa_getkeytab:
    path: /etc/ipa/test.keytab
    principal: HTTP/freeipa-dc02.ipa.test
    ipa_host: freeipa-dc01.ipa.test

- name: Retrieve already existing keytab
  community.general.ipa_getkeytab:
    path: /etc/ipa/test.keytab
    principal: HTTP/freeipa-dc02.ipa.test
    ipa_host: freeipa-dc01.ipa.test
    retrieve_mode: true

- name: Force keytab recreation
  community.general.ipa_getkeytab:
    path: /etc/ipa/test.keytab
    principal: HTTP/freeipa-dc02.ipa.test
    ipa_host: freeipa-dc01.ipa.test
    force: true
"""

import os

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt


class IPAKeytab:
    def __init__(self, module, **kwargs):
        self.module = module
        self.path = kwargs["path"]
        self.state = kwargs["state"]
        self.principal = kwargs["principal"]
        self.ipa_host = kwargs["ipa_host"]
        self.ldap_uri = kwargs["ldap_uri"]
        self.bind_dn = kwargs["bind_dn"]
        self.bind_pw = kwargs["bind_pw"]
        self.password = kwargs["password"]
        self.ca_cert = kwargs["ca_cert"]
        self.sasl_mech = kwargs["sasl_mech"]
        self.retrieve_mode = kwargs["retrieve_mode"]
        self.encryption_types = kwargs["encryption_types"]

        self.runner = CmdRunner(
            module,
            command="ipa-getkeytab",
            arg_formats=dict(
                retrieve_mode=cmd_runner_fmt.as_bool("--retrieve"),
                path=cmd_runner_fmt.as_opt_val("--keytab"),
                ipa_host=cmd_runner_fmt.as_opt_val("--server"),
                principal=cmd_runner_fmt.as_opt_val("--principal"),
                ldap_uri=cmd_runner_fmt.as_opt_val("--ldapuri"),
                bind_dn=cmd_runner_fmt.as_opt_val("--binddn"),
                bind_pw=cmd_runner_fmt.as_opt_val("--bindpw"),
                password=cmd_runner_fmt.as_opt_val("--password"),
                ca_cert=cmd_runner_fmt.as_opt_val("--cacert"),
                sasl_mech=cmd_runner_fmt.as_opt_val("--mech"),
                encryption_types=cmd_runner_fmt.as_opt_val("--enctypes"),
            ),
        )

    def _exec(self, check_rc=True):
        with self.runner(
            "retrieve_mode path ipa_host principal ldap_uri bind_dn bind_pw password ca_cert sasl_mech encryption_types",
            check_rc=check_rc,
        ) as ctx:
            rc, out, err = ctx.run()
        return out


def main():
    arg_spec = dict(
        path=dict(type="path", required=True, aliases=["keytab"]),
        state=dict(default="present", choices=["present", "absent"]),
        principal=dict(type="str", required=True),
        ipa_host=dict(type="str"),
        ldap_uri=dict(type="str"),
        bind_dn=dict(type="str"),
        bind_pw=dict(type="str"),
        password=dict(type="str", no_log=True),
        ca_cert=dict(type="path"),
        sasl_mech=dict(type="str", choices=["GSSAPI", "EXTERNAL"]),
        retrieve_mode=dict(type="bool"),
        encryption_types=dict(type="str"),
        force=dict(type="bool"),
    )
    module = AnsibleModule(
        argument_spec=arg_spec,
        mutually_exclusive=[("ipa_host", "ldap_uri"), ("retrieve_mode", "password")],
        supports_check_mode=True,
    )

    path = module.params["path"]
    state = module.params["state"]
    force = module.params["force"]

    keytab = IPAKeytab(
        module,
        path=path,
        state=state,
        principal=module.params["principal"],
        ipa_host=module.params["ipa_host"],
        ldap_uri=module.params["ldap_uri"],
        bind_dn=module.params["bind_dn"],
        bind_pw=module.params["bind_pw"],
        password=module.params["password"],
        ca_cert=module.params["ca_cert"],
        sasl_mech=module.params["sasl_mech"],
        retrieve_mode=module.params["retrieve_mode"],
        encryption_types=module.params["encryption_types"],
    )

    changed = False
    if state == "present":
        if os.path.exists(path):
            if force and not module.check_mode:
                try:
                    os.remove(path)
                except OSError as e:
                    module.fail_json(msg=f"Error deleting: {e.filename} - {e.strerror}.")
                keytab._exec()
                changed = True
            if force and module.check_mode:
                changed = True
        else:
            changed = True
            keytab._exec()

    if state == "absent":
        if os.path.exists(path):
            changed = True
            if not module.check_mode:
                try:
                    os.remove(path)
                except OSError as e:
                    module.fail_json(msg=f"Error deleting: {e.filename} - {e.strerror}.")

    module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2020 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:

    # Standard files documentation fragment
    DOCUMENTATION = r"""
requirements:
  - lxml

options:
  hosttech_username:
    description:
      - The username for the Hosttech API user.
      - If provided, O(hosttech_password) must also be provided.
      - Mutually exclusive with O(hosttech_token).
    type: str
  hosttech_password:
    description:
      - The password for the Hosttech API user.
      - If provided, O(hosttech_username) must also be provided.
      - Mutually exclusive with O(hosttech_token).
    type: str
  hosttech_token:
    description:
      - The password for the Hosttech API user.
      - Mutually exclusive with O(hosttech_username) and O(hosttech_password).
      - Since community.dns 1.2.0, the alias O(ignore:api_token) can be used.
    aliases:
      - api_token
    type: str
    version_added: 0.2.0
"""

    # NOTE: This document fragment augments the above standard DOCUMENTATION document fragment
    #       by providing alternative ways to provide configuration for plugins. (The above
    #       documentation fragment is tailored for modules.)
    PLUGIN = r"""
options:
  hosttech_username:
    env:
      - name: ANSIBLE_HOSTTECH_API_USERNAME
        version_added: 2.5.0
  hosttech_password:
    env:
      - name: ANSIBLE_HOSTTECH_API_PASSWORD
        version_added: 2.5.0
  hosttech_token:
    env:
      - name: ANSIBLE_HOSTTECH_DNS_TOKEN
        version_added: 2.5.0
"""

    # NOTE: This document fragment adds additional information on records.
    RECORD_NOTES = r"""
options: {}
"""

    RECORD_TYPE_SEEALSO = r"""
options: {}
seealso:
  - ref: ansible_collections.community.dns.docsite.hosttech_guide.records
    description: More information on supported DNS record types.
"""

    # WARNING: This section is automatically generated by update-docs-fragments.py.
    #          It is used to augment the docs fragments module_record, module_record_set.
    #          DO NOT EDIT MANUALLY!
    RECORD_DEFAULT_TTL = r"""
options:
  ttl:
    default: 3600
"""

    # WARNING: This section is automatically generated by update-docs-fragments.py.
    #          It is used to augment the docs fragments module_record, module_record_info,
    #          module_record_set, module_record_set_info.
    #          DO NOT EDIT MANUALLY!
    RECORD_TYPE_CHOICES = r"""
options:
  type:
    choices:
      - A
      - AAAA
      - CAA
      - CNAME
      - MX
      - NS
      - PTR
      - SPF
      - SRV
      - TXT
"""

    # WARNING: This section is automatically generated by update-docs-fragments.py.
    #          It is used to augment the docs fragment module_record_sets.
    #          DO NOT EDIT MANUALLY!
    RECORD_TYPE_CHOICES_RECORD_SETS_MODULE = r"""
options:
  record_sets:
    suboptions:
      record:
        description:
          - The full DNS record to create or delete.
          - Exactly one of O(record_sets[].record) and O(record_sets[].prefix) must
            be specified.
        type: str
      prefix:
        description:
          - The prefix of the DNS record.
          - This is the part of O(record_sets[].record) before O(zone_name). For example,
            if the record to be modified is C(www.example.com) for the zone C(example.com),
            the prefix is V(www). If the record in this example would be C(example.com),
            the prefix would be V('') (empty string).
          - Exactly one of O(record_sets[].record) and O(record_sets[].prefix) must
            be specified.
        type: str
      ttl:
        description:
          - The TTL to give the new record, in seconds.
        type: int
        default: 3600
      type:
        description:
          - The type of DNS record to create or delete.
        required: true
        type: str
        choices:
          - A
          - AAAA
          - CAA
          - CNAME
          - MX
          - NS
          - PTR
          - SPF
          - SRV
          - TXT
      value:
        description:
          - The new value when creating a DNS record.
          - YAML lists or multiple comma-spaced values are allowed.
          - When deleting a record all values for the record must be specified or
            it will not be deleted.
          - Must be specified if O(record_sets[].ignore=false).
        type: list
        elements: str
      ignore:
        description:
          - If set to V(true), O(record_sets[].value) will be ignored.
          - This is useful when O(prune=true), but you do not want certain entries
            to be removed without having to know their current value.
        type: bool
        default: false
"""

    # WARNING: This section is automatically generated by update-docs-fragments.py.
    #          It is used to augment the docs fragment inventory_records.
    #          DO NOT EDIT MANUALLY!
    RECORD_TYPE_CHOICES_RECORDS_INVENTORY = r"""
options:
  simple_filters:
    suboptions:
      type:
        description:
          - Record types whose values to use.
        type: list
        elements: string
        default:
          - A
          - AAAA
          - CNAME
        choices:
          - A
          - AAAA
          - CAA
          - CNAME
          - MX
          - NS
          - PTR
          - SPF
          - SRV
          - TXT
"""

    # WARNING: This section is automatically generated by update-docs-fragments.py.
    #          It is used to augment the docs fragments inventory_records, module_record,
    #          module_record_info, module_record_set, module_record_set_info,
    #          module_record_sets, module_zone_info.
    #          DO NOT EDIT MANUALLY!
    ZONE_ID_TYPE = r"""
options:
  zone_id:
    type: int
"""



# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: machinectl
short_description: Systemd's machinectl privilege escalation
description:
  - This become plugins allows your remote/login user to execute commands as another user using the C(machinectl) utility.
author: Ansible Core Team
options:
  become_user:
    description: User you 'become' to execute the task.
    type: string
    default: ''
    ini:
      - section: privilege_escalation
        key: become_user
      - section: machinectl_become_plugin
        key: user
    vars:
      - name: ansible_become_user
      - name: ansible_machinectl_user
    env:
      - name: ANSIBLE_BECOME_USER
      - name: ANSIBLE_MACHINECTL_USER
  become_exe:
    description: C(machinectl) executable.
    type: string
    default: machinectl
    ini:
      - section: privilege_escalation
        key: become_exe
      - section: machinectl_become_plugin
        key: executable
    vars:
      - name: ansible_become_exe
      - name: ansible_machinectl_exe
    env:
      - name: ANSIBLE_BECOME_EXE
      - name: ANSIBLE_MACHINECTL_EXE
  become_flags:
    description: Options to pass to C(machinectl).
    type: string
    default: ''
    ini:
      - section: privilege_escalation
        key: become_flags
      - section: machinectl_become_plugin
        key: flags
    vars:
      - name: ansible_become_flags
      - name: ansible_machinectl_flags
    env:
      - name: ANSIBLE_BECOME_FLAGS
      - name: ANSIBLE_MACHINECTL_FLAGS
  become_pass:
    description: Password for C(machinectl).
    type: string
    required: false
    vars:
      - name: ansible_become_password
      - name: ansible_become_pass
      - name: ansible_machinectl_pass
    env:
      - name: ANSIBLE_BECOME_PASS
      - name: ANSIBLE_MACHINECTL_PASS
    ini:
      - section: machinectl_become_plugin
        key: password
notes:
  - When not using this plugin with user V(root), it only works correctly with a polkit rule which alters the behaviour
    of C(machinectl). This rule must alter the prompt behaviour to ask directly for the user credentials, if the user is allowed
    to perform the action (take a look at the examples section). If such a rule is not present the plugin only works if it
    is used in context with the root user, because then no further prompt is shown by C(machinectl).
  - This become plugin does not work when connection pipelining is enabled. With ansible-core 2.19+, using it automatically
    disables pipelining. On ansible-core 2.18 and before, pipelining must explicitly be disabled by the user.
"""

EXAMPLES = r"""
# A polkit rule needed to use the module with a non-root user.
# See the Notes section for details.
/etc/polkit-1/rules.d/60-machinectl-fast-user-auth.rules: |-
  polkit.addRule(function(action, subject) {
    if(action.id == "org.freedesktop.machine1.host-shell" &&
      subject.isInGroup("wheel")) {
        return polkit.Result.AUTH_SELF_KEEP;
    }
  });
"""

from re import compile as re_compile

from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins.become import BecomeBase

ansi_color_codes = re_compile(to_bytes(r"\x1B\[[0-9;]+m"))


class BecomeModule(BecomeBase):
    name = "community.general.machinectl"

    prompt = "Password: "
    fail = ("==== AUTHENTICATION FAILED ====",)
    success = ("==== AUTHENTICATION COMPLETE ====",)
    require_tty = True  # see https://github.com/ansible-collections/community.general/issues/6932

    # See https://github.com/ansible/ansible/issues/81254,
    # https://github.com/ansible/ansible/pull/78111
    pipelining = False

    @staticmethod
    def remove_ansi_codes(line):
        return ansi_color_codes.sub(b"", line)

    def build_become_command(self, cmd, shell):
        super().build_become_command(cmd, shell)

        if not cmd:
            return cmd

        become = self.get_option("become_exe")

        flags = self.get_option("become_flags")
        user = self.get_option("become_user")
        # SYSTEMD_COLORS=0 stops machinectl from appending ANSI reset
        # sequences (ESC[0m, ESC[J) after the child exits, which would
        # otherwise land after the module JSON and break result parsing.
        return f"SYSTEMD_COLORS=0 {become} -q shell {flags} {user}@ {self._build_success_command(cmd, shell)}"

    def check_success(self, b_output):
        b_output = self.remove_ansi_codes(b_output)
        return super().check_success(b_output)

    def check_incorrect_password(self, b_output):
        b_output = self.remove_ansi_codes(b_output)
        return super().check_incorrect_password(b_output)

    def check_missing_password(self, b_output):
        b_output = self.remove_ansi_codes(b_output)
        return super().check_missing_password(b_output)



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Bfd_templates parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class Bfd_templatesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Bfd_templatesTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "name",
            "getval": re.compile(
                r"""^bfd-template
                    \s(?P<hop>single-hop|multi-hop)
                    \s(?P<name>\S+)
                    $""",
                re.VERBOSE,
            ),
            "compval": "name",
            "setval": "bfd-template {% if hop == 'single_hop' %}single-hop{% elif hop == 'multi_hop' %}multi-hop{% endif %} {{ name }}",
            "result": {
                "{{ name }}": {
                    "name": "{{ name }}",
                    "hop": "{% if hop == 'single-hop' %}single_hop{% elif hop == 'multi-hop' %}multi_hop{% endif %}",
                },
            },
            "shared": True,
        },
        {
            "name": "interval",
            "getval": re.compile(
                r"""
                \s+interval\smin-tx\s(?P<min_tx>\d+)
                \smin-rx\s(?P<min_rx>\d+)
                \smultiplier\s(?P<multiplier>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "interval min-tx {{ interval.min_tx }} "
                      "min-rx {{ interval.min_rx }} "
                      "multiplier {{ interval.multiplier }}",
            "result": {
                "{{ name }}": {
                    "interval": {
                        "min_tx": "{{ min_tx | int }}",
                        "min_rx": "{{ min_rx | int }}",
                        "multiplier": "{{ multiplier | int }}",
                    },
                },
            },
        },
        {
            "name": "dampening",
            "getval": re.compile(
                r"""
                \s+dampening\s(?P<half_life_period>\d+)
                \s(?P<reuse_threshold>\d+)
                \s(?P<suppress_threshold>\d+)
                \s(?P<max_suppress_time>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "dampening {{ dampening.half_life_period }} "
                      "{{ dampening.reuse_threshold }} "
                      "{{ dampening.suppress_threshold }} "
                      "{{ dampening.max_suppress_time }}",
            "result": {
                "{{ name }}": {
                    "dampening": {
                        "half_life_period": "{{ half_life_period | int }}",
                        "reuse_threshold": "{{ reuse_threshold | int }}",
                        "suppress_threshold": "{{ suppress_threshold | int }}",
                        "max_suppress_time": "{{ max_suppress_time | int }}",
                    },
                },
            },
        },
        {
            "name": "echo",
            "getval": re.compile(
                r"""
                \s+(?P<no>no\s)?echo
                $""",
                re.VERBOSE,
            ),
            "setval": "echo",
            "result": {
                "{{ name }}": {
                    "echo": "{{ False if no is defined else True }}",
                },
            },
        },
        {
            "name": "authentication",
            "getval": re.compile(
                r"""
                \s+authentication\s(?P<auth_type>sha-1|md5|meticulous-md5|meticulous-sha-1)
                \skeychain\s(?P<keychain>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": (
                "authentication "
                "{% if authentication.type == 'sha_1' %}sha-1"
                "{% elif authentication.type == 'md5' %}md5"
                "{% elif authentication.type == 'meticulous_md5' %}meticulous-md5"
                "{% elif authentication.type == 'meticulous_sha_1' %}meticulous-sha-1"
                "{% endif %} "
                "keychain {{ authentication.keychain }}"
            ),
            "result": {
                "{{ name }}": {
                    "authentication": {
                        "type": (
                            "{% if auth_type == 'sha-1' %}sha_1"
                            "{% elif auth_type == 'md5' %}md5"
                            "{% elif auth_type == 'meticulous-md5' %}meticulous_md5"
                            "{% elif auth_type == 'meticulous-sha-1' %}meticulous_sha_1"
                            "{% endif %}"
                        ),
                        "keychain": "{{ keychain }}",
                    },
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_vol
version_added: 1.0.0
short_description: Create and attach a volume, return volume ID and device map
description:
  - Creates an EBS volume and optionally attaches it to an instance.
  - If both O(instance) and O(name) are given and the instance has a device at the device name, then no volume is created and no attachment is made.
options:
  instance:
    description:
      - Instance ID if you wish to attach the volume.
      - Set to V(None) to detach the volume.
    type: str
  name:
    description:
      - Volume Name tag if you wish to attach an existing volume (requires instance).
    type: str
  id:
    description:
      - Volume ID if you wish to attach an existing volume (requires instance) or remove an existing volume.
    type: str
  volume_size:
    description:
      - Size of volume (in GiB) to create.
    type: int
  volume_type:
    description:
      - Type of EBS volume; V(standard) (magnetic), V(gp2) (SSD), V(gp3) (SSD), V(io1) (Provisioned IOPS), V(io2) (Provisioned IOPS),
        V(st1) (Throughput Optimized HDD), V(sc1) (Cold HDD).
      - V(standard) is the old EBS default and continues to remain the Ansible default for backwards compatibility.
    default: standard
    choices: ['standard', 'gp2', 'io1', 'st1', 'sc1', 'gp3', 'io2']
    type: str
  iops:
    description:
      - The provisioned IOPs you want to associate with this volume (integer).
    type: int
  encrypted:
    description:
      - Enable encryption at rest for this volume.
    default: false
    type: bool
  kms_key_id:
    description:
      - Specify the ID of the KMS key to use.
    type: str
  device_name:
    description:
      - Device ID to override device mapping. Assumes /dev/sdf for Linux/UNIX and /dev/xvdf for Windows.
    type: str
  delete_on_termination:
    description:
      - When set to C(true), the volume will be deleted upon instance termination.
    type: bool
    default: false
  zone:
    description:
      - Zone in which to create the volume, if unset uses the zone the instance is in (if set).
    aliases: ['availability_zone', 'aws_zone', 'ec2_zone']
    type: str
  snapshot:
    description:
      - Snapshot ID on which to base the volume.
    type: str
  volume_initialization_rate:
    description:
      - Rate at which to initiliaze the volume from a snapshot.
    type: int
    version_added: 11.2.0
  state:
    description:
      - Whether to ensure the volume is present or absent.
      - O(state=list) was deprecated in release 1.1.0 and is no longer available
        with release 4.0.0.
      - The V(list) functionality has been moved to a dedicated module M(amazon.aws.ec2_vol_info).
    default: present
    choices: ['absent', 'present']
    type: str
  modify_volume:
    description:
      - The volume won't be modified unless this key is V(true).
    type: bool
    default: false
    version_added: 1.4.0
  throughput:
    description:
      - Volume throughput in MB/s.
      - This parameter is only valid for gp3 volumes.
      - Valid range is from 125 to 2000.
    type: int
    version_added: 1.4.0
  multi_attach:
    description:
      - If set to V(true), Multi-Attach will be enabled when creating the volume.
      - When you create a new volume, Multi-Attach is disabled by default.
      - This parameter is supported with io1 and io2 volumes only.
    type: bool
    version_added: 2.0.0
  outpost_arn:
    description:
      - The Amazon Resource Name (ARN) of the Outpost.
      - If set, allows to create volume in an Outpost.
    type: str
    version_added: 3.1.0
author:
  - "Lester Wade (@lwade)"
notes:
  - Support for O(purge_tags) was added in release 1.5.0.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Simple attachment action
- amazon.aws.ec2_vol:
    instance: XXXXXX
    volume_size: 5
    device_name: sdd
    region: us-west-2

# Example using custom iops params
- amazon.aws.ec2_vol:
    instance: XXXXXX
    volume_size: 5
    iops: 100
    device_name: sdd
    region: us-west-2

# Example using snapshot id
- amazon.aws.ec2_vol:
    instance: XXXXXX
    snapshot: "{{ snapshot }}"

# Playbook example combined with instance launch
- amazon.aws.ec2:
    keypair: "{{ keypair }}"
    image: "{{ image }}"
    wait: true
    count: 3
  register: ec2
- amazon.aws.ec2_vol:
    instance: "{{ item.id }}"
    volume_size: 5
  loop: "{{ ec2.instances }}"
  register: ec2_vol

# Example: Launch an instance and then add a volume if not already attached
#   * Volume will be created with the given name if not already created.
#   * Nothing will happen if the volume is already attached.

- amazon.aws.ec2:
    keypair: "{{ keypair }}"
    image: "{{ image }}"
    zone: YYYYYY
    id: my_instance
    wait: true
    count: 1
  register: ec2

- amazon.aws.ec2_vol:
    instance: "{{ item.id }}"
    name: my_existing_volume_Name_tag
    device_name: /dev/xvdf
  loop: "{{ ec2.instances }}"
  register: ec2_vol

# Remove a volume
- amazon.aws.ec2_vol:
    id: vol-XXXXXXXX
    state: absent

# Detach a volume (since 1.9)
- amazon.aws.ec2_vol:
    id: vol-XXXXXXXX
    instance: None
    region: us-west-2

# Create new volume using SSD storage
- amazon.aws.ec2_vol:
    instance: XXXXXX
    volume_size: 50
    volume_type: gp2
    device_name: /dev/xvdf

# Create new volume with multi-attach enabled
- amazon.aws.ec2_vol:
    zone: XXXXXX
    multi_attach: true
    volume_size: 4
    volume_type: io1
    iops: 102

# Attach an existing volume to instance. The volume will be deleted upon instance termination.
- amazon.aws.ec2_vol:
    instance: XXXXXX
    id: XXXXXX
    device_name: /dev/sdf
    delete_on_termination: true
"""

RETURN = r"""
device:
    description: Device name of attached volume.
    returned: when success
    type: str
    sample: "/dev/sdf"
volume_id:
    description: The id of volume.
    returned: when success
    type: str
    sample: "vol-35b333d9"
volume_type:
    description: The volume type.
    returned: when success
    type: str
    sample: "standard"
volume:
    description: A dictionary containing detailed attributes of the volume.
    returned: when success
    type: dict
    contains:
        attachment_set:
            description:
                - Information about the volume attachments.
                - This was changed in version 2.0.0 from a dictionary to a list of dictionaries.
            type: list
            elements: dict
            returned: when success
            sample: [{
                "attach_time": "2015-10-23T00:22:29.000Z",
                "deleteOnTermination": "false",
                "device": "/dev/sdf",
                "instance_id": "i-8356263c",
                "status": "attached"
            }]
        create_time:
            description: The time stamp when volume creation was initiated.
            type: str
            returned: when success
            sample: "2015-10-21T14:36:08.870Z"
        encrypted:
            description: Indicates whether the volume is encrypted.
            type: bool
            returned: when success
            sample: False
        id:
            description: The ID of the volume.
            type: str
            returned: when success
            sample: "vol-35b333d9"
        iops:
            description: The number of I/O operations per second (IOPS) that the volume supports.
            type: int
            returned: when success
            sample: null
        size:
            description: The size of the volume, in GiBs.
            type: int
            returned: when success
            sample: 1
        snapshot_id:
            description: The snapshot from which the volume was created, if applicable.
            type: str
            returned: when success
            sample: ""
        status:
            description: The volume state.
            type: str
            returned: when success
            sample: "in-use"
        tags:
            description: Any tags assigned to the volume.
            type: dict
            returned: when success
            sample: {
                env: "dev"
                }
        type:
            description: The volume type. This can be gp2, io1, st1, sc1, or standard.
            type: str
            returned: when success
            sample: "standard"
        zone:
            description: The Availability Zone of the volume.
            type: str
            returned: when success
            sample: "us-east-1b"
        throughput:
            description: The throughput that the volume supports, in MiB/s.
            type: int
            returned: when success
            sample: 131
"""

import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.arn import is_outpost_arn
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import attach_volume as attach_ec2_volume
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_volume as create_ec2_volume
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_volume as delete_ec2_volume
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_instances
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_volumes
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import detach_volume as detach_ec2_volume
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import modify_instance_attribute
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import modify_volume
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list
from ansible_collections.amazon.aws.plugins.module_utils.waiters import wait_for_resource_state


def get_instance(module: AnsibleAWSModule, ec2_conn, instance_id: Optional[str] = None) -> Optional[Dict[str, Any]]:
    instance = None
    if not instance_id:
        return instance

    try:
        reservation_response = describe_instances(ec2_conn, InstanceIds=[instance_id])
        if reservation_response:
            instance = camel_dict_to_snake_dict(reservation_response[0]["Instances"][0])
    except AnsibleEC2Error as e:
        module.fail_json_aws(e, msg=f"Error while getting instance_id with id {instance_id}")

    return instance


def get_volume(module, ec2_conn, vol_id=None, fail_on_not_found=True):
    name = module.params.get("name")
    param_id = module.params.get("id")
    zone = module.params.get("zone")

    if not vol_id:
        vol_id = param_id

    # If no name or id supplied, just try volume creation based on module parameters
    if vol_id is None and name is None:
        return None

    find_params = dict()
    vols = []

    if vol_id:
        find_params["VolumeIds"] = [vol_id]
    elif name:
        find_params["Filters"] = ansible_dict_to_boto3_filter_list({"tag:Name": name})
    elif zone:
        find_params["Filters"] = ansible_dict_to_boto3_filter_list({"availability-zone": zone})

    try:
        vols = describe_volumes(ec2_conn, **find_params)
        if not vols:
            if fail_on_not_found and vol_id:
                msg = f"Could not find volume with id: {vol_id}"
                if name:
                    msg += f" and name: {name}"
                module.fail_json(msg=msg)
            else:
                return None
    except AnsibleEC2Error as e:
        module.fail_json_aws(e, msg=f"Error while getting EBS volumes with the parameters {find_params}")

    if len(vols) > 1:
        module.fail_json(
            msg=f"Found more than one volume in zone (if specified) with name: {name}",
            found=[v["VolumeId"] for v in vols],
        )
    vol = camel_dict_to_snake_dict(vols[0])
    return vol


def delete_volume(module: AnsibleAWSModule, ec2_conn, volume_id: Optional[str] = None) -> bool:
    changed = False
    if volume_id:
        try:
            changed = delete_ec2_volume(ec2_conn, volume_id)
        except AnsibleEC2Error as e:
            module.fail_json_aws(e, msg="Error while deleting volume")
        wait_for_resource_state(ec2_conn, module, "volume_deleted", VolumeIds=[volume_id])
    return changed


def update_volume(module: AnsibleAWSModule, ec2_conn, volume: Dict[str, Any]) -> Tuple[Dict[str, Any], bool]:
    changed = False
    req_obj = {"VolumeId": volume["volume_id"]}

    if module.params.get("modify_volume"):
        target_type = module.params.get("volume_type")
        original_type = None
        type_changed = False
        if target_type:
            original_type = volume["volume_type"]
            if target_type != original_type:
                type_changed = True
                req_obj["VolumeType"] = target_type

        iops_changed = False
        target_iops = module.params.get("iops")
        original_iops = volume.get("iops")
        if target_iops:
            if target_iops != original_iops:
                iops_changed = True
                req_obj["Iops"] = target_iops
            else:
                req_obj["Iops"] = original_iops
        else:
            # If no IOPS value is specified and there was a volume_type update to gp3,
            # the existing value is retained, unless a volume type is modified that supports different values,
            # otherwise, the default iops value is applied.
            if type_changed and target_type == "gp3":
                if (original_iops and (int(original_iops) < 3000 or int(original_iops) > 80000)) or not original_iops:
                    req_obj["Iops"] = 3000
                    iops_changed = True

        target_size = module.params.get("volume_size")
        size_changed = False
        if target_size:
            original_size = volume["size"]
            if target_size != original_size:
                size_changed = True
                req_obj["Size"] = target_size

        target_type = module.params.get("volume_type")
        original_type = None
        type_changed = False
        if target_type:
            original_type = volume["volume_type"]
            if target_type != original_type:
                type_changed = True
                req_obj["VolumeType"] = target_type

        target_throughput = module.params.get("throughput")
        throughput_changed = False
        if target_throughput:
            original_throughput = volume.get("throughput")
            if target_throughput != original_throughput:
                throughput_changed = True
                req_obj["Throughput"] = target_throughput

        target_multi_attach = module.params.get("multi_attach")
        multi_attach_changed = False
        if target_multi_attach is not None:
            original_multi_attach = volume["multi_attach_enabled"]
            if target_multi_attach != original_multi_attach:
                multi_attach_changed = True
                req_obj["MultiAttachEnabled"] = target_multi_attach

        changed = iops_changed or size_changed or type_changed or throughput_changed or multi_attach_changed

        if changed:
            if module.check_mode:
                module.exit_json(changed=True, msg="Would have updated volume if not in check mode.")
            try:
                response = modify_volume(ec2_conn, **req_obj)
            except AnsibleEC2Error as e:
                module.fail_json_aws(e, msg="Error while modifying volume")

            volume["size"] = response.get("TargetSize")
            volume["volume_type"] = response.get("TargetVolumeType")
            volume["iops"] = response.get("TargetIops")
            volume["multi_attach_enabled"] = response.get("TargetMultiAttachEnabled")
            volume["throughput"] = response.get("TargetThroughput")

    return volume, changed


def create_volume(module: AnsibleAWSModule, ec2_conn, zone: str) -> Tuple[Dict[str, Any], bool]:
    changed = False
    iops = module.params.get("iops")
    encrypted = module.params.get("encrypted")
    kms_key_id = module.params.get("kms_key_id")
    volume_size = module.params.get("volume_size")
    volume_type = module.params.get("volume_type")
    snapshot = module.params.get("snapshot")
    throughput = module.params.get("throughput")
    multi_attach = module.params.get("multi_attach")
    outpost_arn = module.params.get("outpost_arn")
    tags = module.params.get("tags") or {}
    name = module.params.get("name")
    volume_initialization_rate = module.params.get("volume_initialization_rate")

    volume = get_volume(module, ec2_conn)

    if module.check_mode:
        module.exit_json(changed=True, msg="Would have created a volume if not in check mode.")

    if volume is None:
        changed = True
        additional_params = dict()

        if volume_size:
            additional_params["Size"] = int(volume_size)

        if kms_key_id:
            additional_params["KmsKeyId"] = kms_key_id

        if snapshot:
            additional_params["SnapshotId"] = snapshot

        if volume_initialization_rate:
            additional_params["VolumeInitializationRate"] = int(volume_initialization_rate)

        if iops:
            additional_params["Iops"] = int(iops)

        # Use the default value if any iops has been specified when volume_type=gp3
        if volume_type == "gp3" and not iops:
            additional_params["Iops"] = 3000

        if throughput:
            additional_params["Throughput"] = int(throughput)

        if multi_attach:
            additional_params["MultiAttachEnabled"] = True

        if outpost_arn:
            if is_outpost_arn(outpost_arn):
                additional_params["OutpostArn"] = outpost_arn
            else:
                module.fail_json("OutpostArn does not match the pattern specified in API specifications.")

        if name:
            tags["Name"] = name

        if tags:
            additional_params["TagSpecifications"] = boto3_tag_specifications(tags, types=["volume"])

        try:
            create_vol_response = create_ec2_volume(
                ec2_conn, AvailabilityZone=zone, Encrypted=encrypted, VolumeType=volume_type, **additional_params
            )
        except AnsibleEC2Error as e:
            module.fail_json_aws(e, msg="Error while creating EBS volume")

        wait_for_resource_state(ec2_conn, module, "volume_available", VolumeIds=[create_vol_response["VolumeId"]])
        volume = get_volume(module, ec2_conn, vol_id=create_vol_response["VolumeId"])

    return volume, changed


def attach_volume(
    module: AnsibleAWSModule, ec2_conn, volume_dict: Dict[str, Any], instance_dict: Dict[str, Any], device_name: str
) -> Tuple[Dict[str, Any], bool]:
    changed = False

    # If device_name isn't set, make a choice based on best practices here:
    # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/block-device-mapping-concepts.html

    # In future this needs to be more dynamic but combining block device mapping best practices
    # (bounds for devices, as above) with instance.block_device_mapping data would be tricky. For me ;)

    attachment_data = get_attachment_data(volume_dict, wanted_state="attached")
    if attachment_data:
        if module.check_mode:
            if attachment_data[0].get("status") in ["attached", "attaching"]:
                instance_id = attachment_data[0].get("instance_id", "None")
                module.exit_json(
                    changed=False,
                    msg=f"IN CHECK MODE - volume already attached to instance: {instance_id}.",
                    volume=volume_dict,
                )
        if not volume_dict["multi_attach_enabled"]:
            # volumes without MultiAttach Enabled can be attached to 1 instance only
            if attachment_data[0].get("instance_id", None) != instance_dict["instance_id"]:
                instance_id = attachment_data[0].get("instance_id", "None")
                module.fail_json(
                    msg=f"Volume {volume_dict['volume_id']} is already attached to another instance: {instance_id}."
                )
            else:
                return volume_dict, changed

    try:
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have attached volume if not in check mode.")
        attach_response = attach_ec2_volume(
            ec2_conn, device=device_name, instance_id=instance_dict["instance_id"], volume_id=volume_dict["volume_id"]
        )
    except AnsibleEC2Error as e:
        module.fail_json_aws(e, msg="Error while attaching EBS volume")

    wait_for_resource_state(ec2_conn, module, "volume_in_use", VolumeIds=[attach_response["VolumeId"]])
    changed = True

    modify_dot_attribute(module, ec2_conn, instance_dict, device_name)

    volume = get_volume(module, ec2_conn, vol_id=volume_dict["volume_id"])

    return volume, changed


def modify_dot_attribute(module: AnsibleAWSModule, ec2_conn, instance_dict: Dict[str, Any], device_name: str) -> bool:
    """Modify delete_on_termination attribute"""

    delete_on_termination = module.params.get("delete_on_termination")
    changed = False

    # volume_in_use can return *shortly* before it appears on the instance
    # description
    mapped_block_device = None
    _attempt = 0
    while mapped_block_device is None:
        _attempt += 1
        instance_dict = get_instance(module, ec2_conn=ec2_conn, instance_id=instance_dict["instance_id"])
        mapped_block_device = get_mapped_block_device(instance_dict=instance_dict, device_name=device_name)
        if mapped_block_device is None:
            if _attempt > 2:
                module.fail_json(msg="Unable to find device on instance", device=device_name, instance=instance_dict)
            time.sleep(1)

    if delete_on_termination != mapped_block_device["ebs"].get("delete_on_termination"):
        try:
            modify_instance_attribute(
                ec2_conn,
                instance_id=instance_dict["instance_id"],
                BlockDeviceMappings=[
                    {"DeviceName": device_name, "Ebs": {"DeleteOnTermination": delete_on_termination}}
                ],
            )
            changed = True
        except AnsibleEC2Error as e:
            module.fail_json_aws(
                e, msg=f"Error while modifying Block Device Mapping of instance {instance_dict['instance_id']}"
            )

    return changed


def get_attachment_data(volume_dict: Dict[str, Any], wanted_state: Optional[str] = None) -> List[Dict[str, Any]]:
    attachment_data = []
    if not volume_dict:
        return attachment_data
    resource = volume_dict.get("attachments", [])
    if wanted_state:
        # filter 'state', return attachment matching wanted state
        resource = [data for data in resource if data["state"] == wanted_state]

    for data in resource:
        attachment_data.append(
            {
                "attach_time": data.get("attach_time", None),
                "device": data.get("device", None),
                "instance_id": data.get("instance_id", None),
                "status": data.get("state", None),
                "delete_on_termination": data.get("delete_on_termination", None),
            }
        )

    return attachment_data


def detach_volume(module: AnsibleAWSModule, ec2_conn, volume_dict: Dict[str, Any]) -> Tuple[Dict[str, Any], bool]:
    changed = False

    attachment_data = get_attachment_data(volume_dict, wanted_state="attached")
    # The ID of the instance must be specified if you are detaching a Multi-Attach enabled volume.
    for attachment in attachment_data:
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have detached volume if not in check mode.")
        try:
            detach_ec2_volume(ec2_conn, volume_id=volume_dict["volume_id"], InstanceId=attachment["instance_id"])
        except AnsibleEC2Error as e:
            module.fail_json_aws(e, msg="Error while detaching volume")

        wait_for_resource_state(ec2_conn, module, "volume_available", VolumeIds=[volume_dict["volume_id"]])
        changed = True

    volume_dict = get_volume(module, ec2_conn, vol_id=volume_dict["volume_id"])
    return volume_dict, changed


def get_volume_info(volume: Dict[str, Any], tags: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
    if not tags:
        tags = boto3_tag_list_to_ansible_dict(volume.get("tags"))
    attachment_data = get_attachment_data(volume)
    volume_info = {
        "create_time": volume.get("create_time"),
        "encrypted": volume.get("encrypted"),
        "id": volume.get("volume_id"),
        "iops": volume.get("iops"),
        "size": volume.get("size"),
        "snapshot_id": volume.get("snapshot_id"),
        "status": volume.get("state"),
        "type": volume.get("volume_type"),
        "zone": volume.get("availability_zone"),
        "attachment_set": attachment_data,
        "multi_attach_enabled": volume.get("multi_attach_enabled"),
        "tags": tags,
    }

    volume_info["throughput"] = volume.get("throughput")

    return volume_info


def get_mapped_block_device(
    instance_dict: Optional[Dict[str, Any]] = None, device_name: Optional[str] = None
) -> Optional[str]:
    mapped_block_device = None
    if not instance_dict:
        return mapped_block_device
    if not device_name:
        return mapped_block_device

    for device in instance_dict.get("block_device_mappings", []):
        if device["device_name"] == device_name:
            mapped_block_device = device
            break

    return mapped_block_device


def ensure_tags(
    module: AnsibleAWSModule, connection, res_id: str, res_type: str, tags: Dict[str, Any], purge_tags: bool
) -> Tuple[Dict[str, Any], bool]:
    if module.check_mode:
        return {}, True
    changed = ensure_ec2_tags(connection, module, res_id, res_type, tags, purge_tags, ["InvalidVolume.NotFound"])
    final_tags = describe_ec2_tags(connection, module, res_id, res_type)

    return final_tags, changed


def ensure_present(ec2_conn, module: AnsibleAWSModule, volume: Optional[Dict[str, Any]]) -> None:
    instance = module.params.get("instance")
    name = module.params.get("name")
    zone = module.params.get("zone")
    device_name = module.params.get("device_name")
    tags = module.params.get("tags")

    # Ensure we have the zone or can get the zone
    if instance is None and zone is None:
        module.fail_json(msg="You must specify either instance or zone")

    # Set volume detach flag
    if instance == "None" or instance == "":
        instance = None
        detach_vol_flag = True
    else:
        detach_vol_flag = False

    # Here we need to get the zone info for the instance. This covers situation where
    # instance is specified but zone isn't.
    # Useful for playbooks chaining instance launch with volume create + attach and where the
    # zone doesn't matter to the user.
    inst = None
    if instance:
        inst = get_instance(module, ec2_conn, instance_id=instance)
        zone = inst["placement"]["availability_zone"]

        # Use platform attribute to guess whether the instance is Windows or Linux
        if device_name is None:
            if inst.get("platform", "") == "Windows":
                device_name = "/dev/xvdf"
            else:
                device_name = "/dev/sdf"

        # Check if there is a volume already mounted there.
        mapped_device = get_mapped_block_device(instance_dict=inst, device_name=device_name)
        if mapped_device:
            other_volume_mapped = False
            if volume:
                if volume["volume_id"] != mapped_device["ebs"]["volume_id"]:
                    other_volume_mapped = True
            else:
                # No volume found so this is another volume
                other_volume_mapped = True

            if other_volume_mapped:
                module.exit_json(
                    msg=f"Volume mapping for {device_name} already exists on instance {instance}",
                    volume_id=mapped_device["ebs"]["volume_id"],
                    found_volume=volume,
                    device=device_name,
                    changed=False,
                )
    final_tags = None
    tags_changed = False
    if volume:
        volume, changed = update_volume(module, ec2_conn, volume)
        if name:
            if not tags:
                tags = boto3_tag_list_to_ansible_dict(volume.get("tags"))
            tags["Name"] = name
        final_tags, tags_changed = ensure_tags(
            module, ec2_conn, volume["volume_id"], "volume", tags, module.params.get("purge_tags")
        )
    else:
        volume, changed = create_volume(module, ec2_conn, zone=zone)

    if detach_vol_flag:
        volume, attach_changed = detach_volume(module, ec2_conn, volume_dict=volume)
    elif inst is not None:
        volume, attach_changed = attach_volume(
            module, ec2_conn, volume_dict=volume, instance_dict=inst, device_name=device_name
        )
    else:
        attach_changed = False

    # Add device, volume_id and volume_type parameters separately to maintain backward compatibility
    volume_info = get_volume_info(volume, tags=final_tags)

    if tags_changed or attach_changed:
        changed = True

    module.exit_json(
        changed=changed,
        volume=volume_info,
        device=device_name,
        volume_id=volume_info["id"],
        volume_type=volume_info["type"],
    )


def ensure_absent(ec2_conn, module: AnsibleAWSModule, volume: Optional[Dict[str, Any]]) -> None:
    name = module.params.get("name")
    param_id = module.params.get("id")
    changed = False
    if not name and not param_id:
        module.fail_json("A volume name or id is required for deletion")
    if volume and volume.get("state") not in ("deleting", "deleted"):
        if module.check_mode:
            module.exit_json(changed=True, msg="Would have deleted volume if not in check mode.")
        detach_volume(module, ec2_conn, volume_dict=volume)
        changed = delete_volume(module, ec2_conn, volume_id=volume["volume_id"])
    module.exit_json(changed=changed)


def main():
    argument_spec = dict(
        instance=dict(),
        id=dict(),
        name=dict(),
        volume_size=dict(type="int"),
        volume_type=dict(default="standard", choices=["standard", "gp2", "io1", "st1", "sc1", "gp3", "io2"]),
        iops=dict(type="int"),
        encrypted=dict(default=False, type="bool"),
        kms_key_id=dict(),
        device_name=dict(),
        delete_on_termination=dict(default=False, type="bool"),
        zone=dict(aliases=["availability_zone", "aws_zone", "ec2_zone"]),
        snapshot=dict(),
        state=dict(default="present", choices=["absent", "present"]),
        tags=dict(type="dict", aliases=["resource_tags"]),
        modify_volume=dict(default=False, type="bool"),
        throughput=dict(type="int"),
        outpost_arn=dict(type="str"),
        purge_tags=dict(type="bool", default=True),
        multi_attach=dict(type="bool"),
        volume_initialization_rate=dict(type="int"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_if=[
            ["volume_type", "io1", ["iops"]],
            ["volume_type", "io2", ["iops"]],
        ],
        supports_check_mode=True,
    )

    param_id = module.params.get("id")
    name = module.params.get("name")
    volume_size = module.params.get("volume_size")
    snapshot = module.params.get("snapshot")
    state = module.params.get("state")
    iops = module.params.get("iops")
    volume_type = module.params.get("volume_type")
    throughput = module.params.get("throughput")
    multi_attach = module.params.get("multi_attach")
    volume_initialization_rate = module.params.get("volume_initialization_rate")

    if iops:
        if volume_type in ("gp2", "st1", "sc1", "standard"):
            module.fail_json(msg="IOPS is not supported for gp2, st1, sc1, or standard volumes.")

        if volume_type == "gp3" and (int(iops) < 3000 or int(iops) > 80000):
            module.fail_json(msg="For a gp3 volume type, IOPS values must be between 3000 and 80000.")

        if volume_type in ("io1", "io2") and (int(iops) < 100 or int(iops) > 64000):
            module.fail_json(msg="For io1 and io2 volume types, IOPS values must be between 100 and 64000.")

    if throughput:
        if volume_type != "gp3":
            module.fail_json(msg="Throughput is only supported for gp3 volume.")
        if throughput < 125 or throughput > 2000:
            module.fail_json(msg="Throughput values must be between 125 and 2000.")

    if multi_attach is True and volume_type not in ("io1", "io2"):
        module.fail_json(msg="multi_attach is only supported for io1 and io2 volumes.")

    if volume_initialization_rate:
        if not snapshot:
            module.fail_json(msg="Volume initialization rate is only supported when creating a volume from snapshot.")
        if int(volume_initialization_rate) < 100 or int(volume_initialization_rate) > 300:
            module.fail_json(msg="Volume initialization rate must be between 100 and 300 MiB/s.")

    ec2_conn = module.client("ec2")

    # Delaying the checks until after the instance check allows us to get volume ids for existing volumes
    # without needing to pass an unused volume_size
    if not volume_size and not (param_id or name or snapshot):
        module.fail_json(msg="You must specify volume_size or identify an existing volume by id, name, or snapshot")

    # Try getting volume
    volume = get_volume(module, ec2_conn, fail_on_not_found=False)
    if state == "present":
        ensure_present(ec2_conn, module, volume)
    elif state == "absent":
        ensure_absent(ec2_conn, module, volume)


if __name__ == "__main__":
    main()



# Copyright (c) 2026 Bill Peck, <bpeck@redhat.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import traceback
try:
    from typing import Dict
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
    from azure.ai.ml import MLClient
    from azure.ai.ml._restclient.v2022_02_01_preview.models import ListViewType
    from azure.ai.ml._restclient.v2022_10_01_preview.models import ScheduleListViewType
    from azure.ai.ml.entities._builders.base_node import BaseNode
    from azure.ai.ml._restclient.v2020_09_01_dataplanepreview.models import BatchJobResource
    import json
except ImportError:
    pass


class MLClientCommon(AzureRMModuleBaseExt):

    @property
    def client(self):
        self.log('Getting client')
        if not self._client:
            self._client = MLClient(self.azure_auth.azure_credential_track2,
                                    subscription_id=self.subscription_id,
                                    resource_group_name=self.resource_group,
                                    workspace_name=self.ml_workspace,
                                    registry_name=self.ml_registry)
        return self._client

    def entity_to_dict(self, entity):
        """
        ENTITY._to_dict() returns and OrderedDict so we abuse json
        dumps and loads to return a Dict
        """
        if isinstance(entity, BatchJobResource):
            entity = entity.as_dict()
        if isinstance(entity, BaseNode):
            entity = entity._to_job()  # pylint: disable=protected-access
        if isinstance(entity, Dict):
            return entity
        try:
            entity = json.loads(json.dumps(entity._to_dict()))
            return entity
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.module.warn("Failed to deserialize response: %s" % str(err))
            self.module.warn(str(entity))
            self.module.debug(traceback.format_exc())

    def filter_required(self, workspace):
        """
        Filter out required networks which are automatically
        added so we can be idempotent.
        """

        if "managed_network" in workspace and \
                "outbound_rules" in workspace["managed_network"]:
            outbound_rules = workspace["managed_network"].pop("outbound_rules")
            updated_rules = []
            for outbound_rule in outbound_rules:
                if outbound_rule.get("category") != "required":
                    updated_rules.append(outbound_rule)
            workspace["managed_network"]["outbound_rules"] = updated_rules

        return workspace

    def get_list_view_type(self, list_type):
        list_types = dict(active=ListViewType.ACTIVE_ONLY,
                          archived=ListViewType.ARCHIVED_ONLY,
                          all=ListViewType.ALL)
        return list_types.get(list_type, ListViewType.ACTIVE_ONLY)

    def get_schedule_list_view_type(self, list_type):
        list_types = dict(enabled=ScheduleListViewType.ENABLED_ONLY,
                          disabled=ScheduleListViewType.DISABLED_ONLY,
                          all=ScheduleListViewType.ALL)
        return list_types.get(list_type, ScheduleListViewType.ENABLED_ONLY)
